from .base import GeocodeService
import logging
from omgeo.places import Candidate
from omgeo.preprocessors import ReplaceRangeWithNumber
from posixpath import join as posixjoin

try:
    # python 3
    from urllib.parse import urljoin
except ImportError:
    # python 2
    from urlparse import urljoin

logger = logging.getLogger(__name__)


class Pelias(GeocodeService):
    """
    Class to geocode using `Pelias search service
    <https://pelias.io/>`_.

    Settings used by the Pelias GeocodeService object include:
     * api_key --  The API key used to access search service.

    """
    _wkid = 4326

    # 766-68 Any St. -> 766 Any St.
    DEFAULT_PREPROCESSORS = [ReplaceRangeWithNumber()]
    DEFAULT_POSTPROCESSORS = []

    def __init__(self, preprocessors=None, postprocessors=None, settings=None):
        if 'api_version' in settings:
            self._api_version = 'v' + str(settings['api_version'])
        else:
            self._api_version = 'v1'

        if 'instance_url' in settings:
            self._base_url = settings['instance_url']
        else:
            self._base_url = 'https://api.geocode.earth/'

        self._default_endpoint = urljoin(self._base_url,
                                         posixjoin(self._api_version, 'search'))
        self._key_endpoint = urljoin(self._base_url,
                                     posixjoin(self._api_version, 'place'))
        self._endpoint = self._default_endpoint

        preprocessors = Pelias.DEFAULT_PREPROCESSORS if preprocessors is None else preprocessors
        postprocessors = Pelias.DEFAULT_POSTPROCESSORS if postprocessors is None else postprocessors
        GeocodeService.__init__(self, preprocessors, postprocessors, settings)

    def _geocode(self, pq):
        query = {'text': pq.query}

        if pq.country:
            query = dict(query, **{'boundary.country': pq.country})

        if pq.viewbox is not None:
            box = pq.viewbox.to_pelias_dict()
            query = dict(query, **box)

        if hasattr(pq, 'key'):
            # Swap to the place endpoint and return a single result.
            self._endpoint = self._key_endpoint
            query = {'ids': pq.key}

        if 'api_key' in self._settings:
            query['api_key'] = self._settings['api_key']

        response_obj = self._get_json_obj(self._endpoint, query)
        returned_candidates = []  # this will be the list returned
        features_in_response = response_obj['features']
        for r in features_in_response:
            properties = r['properties']
            geometry = r['geometry']

            score = 100 * float(properties['confidence']) if 'confidence' in properties else 0
            locality = properties['locality'] if 'locality' in properties else ''
            region = properties['region'] if 'region' in properties else ''
            label = properties['label'] if 'label' in properties else ''
            layer = properties['layer'] if 'layer' in properties else ''

            c = Candidate()
            c.locator = layer
            c.match_addr = label
            c.match_region = region
            c.match_city = locality
            c.locator_type = layer
            c.x = float(geometry['coordinates'][0])
            c.y = float(geometry['coordinates'][1])
            c.score = score
            c.wkid = self._wkid
            c.geoservice = self.__class__.__name__
            returned_candidates.append(c)
        return returned_candidates



from omgeo.processor import _Processor
import math
from operator import attrgetter

# Python 2/3 compatibility: set up 'unicode' for use in string type checking
try:
    unicode
except NameError:
    unicode = str


class _PostProcessor(_Processor):
    """Takes, processes, and returns list of geocoding.places.Candidate objects."""
    def process(self, candidates):
        raise NotImplementedError(
            'PostProcessor subclasses must implement process().')

    def is_case_sensitive(self):
        try:
            return 'CS' if self.case_sensitive else 'CI'
        except ValueError:
            return 'NA'

    def is_exact(self):
        try:
            return 'EXACT_MATCH' if self.exact_match else 'INEXACT_MATCH'
        except ValueError:
            return 'NA'


class LocatorFilter(_PostProcessor):
    """
    PostProcessor used to ditch results with lousy locators.

    :arg list good_locators:  A list of locators to accept results from (default [])
    """
    def __init__(self, good_locators):
        """
        :arg list good_locators:  A list of locators to accept results from (default None)
        """
        self.good_locators = good_locators

    def process(self, candidates):
        """
        :arg list candidates: list of Candidate instances
        """
        for c in candidates[:]:
            if c.locator not in self.good_locators:
                # TODO: search string, i.e. find "EU_Street_Name" in "EU_Street_Name.GBR_StreetName"
                candidates.remove(c)

        return candidates

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.good_locators)


class LocatorSorter(_PostProcessor):
    """
    PostProcessor used to sort by locators
    """

    def __init__(self, ordered_locators):
        """
        :arg list ordered_locators: a list of :py:attr:`Candidate.locator` values
                                    placed in the desired order, such as ``rooftop``,
                                    ``interpolation``, or ``postal``.
        """
        self.ordered_locators = ordered_locators

    def process(self, unordered_candidates):
        """
        :arg list candidates: list of Candidate instances

        """
        ordered_candidates = []
        # make a new list of candidates in order of ordered_locators
        for locator in self.ordered_locators:
            for uc in unordered_candidates[:]:
                if uc.locator == locator:
                    ordered_candidates.append(uc)
                    unordered_candidates.remove(uc)
        # add all the candidates that are still left
        # (whose locator values are not in ordered_locators)
        # and return the new list
        return ordered_candidates + unordered_candidates

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.ordered_locators)


class AttrRename(_PostProcessor):
    """
    PostProcessor used to rename the given attribute, with unspecified
    attributes appearing at the end of the list.
    """

    def __init__(self, attr, attr_map=None, exact_match=False, case_sensitive=False):
        """
        :arg str attr: Name of the attribute
        :arg dict attr_map: Map of old names : new names.
        :arg bool exact_match:
        :arg bool case_sensitive:
        """
        self.attr = attr
        self.attr_map = attr_map if attr_map is not None else {}
        self.exact_match = exact_match
        self.case_sensitive = case_sensitive

    def process(self, candidates):
        """
        :arg list candidates: list of Candidate instances
        :returns: list of Candidate instances with modified values for the given attribute
        """
        def _cc(str_):  # change case
            return str_ if self.case_sensitive else str_.lower()

        new_candidates = []
        for c in candidates[:]:
            attr_val = getattr(c, self.attr)
            if self.exact_match is False and any(_cc(k) in _cc(attr_val) for k in self.attr_map):
                map_key = [k for k in self.attr_map if _cc(k) in _cc(attr_val)][0]
                map_val = self.attr_map[map_key]
                setattr(c, self.attr, map_val)
            elif _cc(attr_val) in [_cc(a) for a in self.attr_map]:
                map_key = [k for k in self.attr_map if _cc(k) == _cc(attr_val)][0]
                setattr(c, self.attr, self.attr_map[map_key])
            new_candidates.append(c)
        return new_candidates

    def __repr__(self):
        return '<%s: %s %s Map of %s (old:new): %s>' \
            % (self.__class__.__name__, self.is_exact(), self.is_case_sensitive(), self.attr, self.attr_map)


class UseHighScoreIfAtLeast(_PostProcessor):
    """
    Limit results to results with at least the given score,
    if and only if one or more results has, at least, the
    given score. If no results have at least this score,
    all of the original results are returned intact.
    """
    def __init__(self, min_score):
        self.min_score = min_score

    def process(self, candidates):
        """
        :arg list candidates: list of Candidates
        :returns: list of Candidates where score is at least min_score,
                  if and only if one or more Candidates have at least min_score.
                  Otherwise, returns original list of Candidates.
        """
        high_score_candidates = [c for c in candidates if c.score >= self.min_score]
        if high_score_candidates != []:
            return high_score_candidates
        return candidates

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.min_score)


class ScoreSorter(_PostProcessor):
    """PostProcessor class to sort :py:class:`Candidate` scores."""
    def __init__(self, reverse=True):
        """
        :arg bool reverse: indicates if the scores should be sorted in descending
                           order (e.g. 100, 90, 80, ...) (default ``True``)
        """
        self.reverse = reverse

    def process(self, candidates):
        """
        :arg list candidates: list of Candidates
        :returns: score-sorted list of Candidates
        """
        return sorted(candidates, key=attrgetter('score'), reverse=self.reverse)

    def __repr__(self):
        order = 'high to low' if self.reverse else 'low to high'
        return '<%s: %s>' % (self.__class__.__name__, order)


class AttrSorter(_PostProcessor):
    """
    PostProcessor used to sort by a the given attribute, with unspecified
    attributes appearing at the end of the list.

    :arg list ordered_values: A list of values placed in the desired order.
    :arg str attr: The attribute on which to sort.
    """

    def __init__(self, ordered_values=None, attr='locator'):
        self.ordered_values = [] if ordered_values is None else ordered_values
        self.attr = attr

    def process(self, unordered_candidates):
        ordered_candidates = []
        # make a new list of candidates in order of ordered_values
        for value in self.ordered_values:
            for uc in unordered_candidates[:]:
                if getattr(uc, self.attr) == value:
                    ordered_candidates.append(uc)
                    unordered_candidates.remove(uc)
        # add all the candidates that are still left
        # and return the new list
        return ordered_candidates + unordered_candidates

    def __repr__(self):
        return '<%s: %s sorted by %s>' % \
            (self.__class__.__name__, self.attr, self.ordered_values)


class AttrReverseSorter(_PostProcessor):
    """
    PostProcessor used to sort by the given attributes in reverse order,
    with unspecified attributes appearing at the end of the list.

    This is good to use when a list has already been defined in a script
    and you are too lazy to use the reverse() function, or don't want
    to in order to maintain more readable code.
    """

    def __init__(self, ordered_values=None, attr='locator'):
        """
        :arg list ordered_values: A list of values placed in the reverse of the desired order.
        :arg str attribute: The attribute on which to sort
        """
        self.ordered_values = [] if ordered_values is None else ordered_values
        self.attr = attr

    def process(self, unordered_candidates):
        ordered_values = self.ordered_values
        ordered_values.reverse()
        sorter = AttrSorter(ordered_values, self.attr)
        return sorter.process(unordered_candidates)

    def __repr__(self):
        return '<%s: %s reverse sorted by %s>' % \
            (self.__class__.__name__, self.attr, self.ordered_values)


class AttrMigrator(_PostProcessor):
    """
    PostProcessor used to migrate the given attribute
    to another attribute.
    """
    def __init__(self, attr_from, attr_to, attr_map=None, exact_match=False, case_sensitive=False):
        self.attr_from = attr_from
        self.attr_to = attr_to
        self.attr_map = {} if attr_map is None else attr_map
        self.exact_match = exact_match
        self.case_sensitive = case_sensitive

    def process(self, candidates):
        def _cc(str_):  # change case
            if self.case_sensitive is False:
                return str_.lower()
            return str_

        new_candidates = []
        for c in candidates[:]:
            from_val = getattr(c, self.attr_from)
            if self.exact_match is False and any(_cc(k) in _cc(from_val) for k in self.attr_map):
                map_key = [k for k in self.attr_map if _cc(k) in _cc(from_val)][0]
                map_val = self.attr_map[map_key]
                setattr(c, self.attr_to, map_val)
            elif _cc(from_val) in [_cc(a) for a in self.attr_map]:
                map_key = [k for k in self.attr_map if _cc(k) == _cc(from_val)][0]
                setattr(c, self.attr_to, self.attr_map[map_key])
            new_candidates.append(c)
        return new_candidates

    def __repr__(self):
        return '<%s: %s -> %s %s %s>' % \
            (self.__class__.__name__, self.attr_from, self.attr_to, self.is_exact(), self.is_case_sensitive())


class AttrFilter(_PostProcessor):
    """
    PostProcessor used to filter out results without desired attribute values.
    """

    def __init__(self, good_values=[], attr='locator', exact_match=True):
        """
        :arg list good_values: A list of values whose candidates we will
                               accept results from (default [])
        :arg string attr: The attribute type on which to filter
        :arg bool exact_match: True if attribute must match a good value exactly.
                               False if the attribute can be a substring in a
                               good value. In other words, if our Candidate
                               attribute is 'US_Rooftop' and one of the good_values
                               is 'Rooftop', we will keep this candidate.
        """
        self._init_helper(vars())

    def process(self, candidates):
        if self.exact_match is True:
            return [c for c in candidates if getattr(c, self.attr) in self.good_values]
        else:
            return [c for c in candidates if any(gv in getattr(c, self.attr)
                                                 for gv in self.good_values)]

    def __repr__(self):
        return '<%s: %s %s in %s>' % \
            (self.__class__.__name__, self.is_exact(), self.attr, self.good_values)


class AttrExclude(_PostProcessor):
    """
    PostProcessor used to filter out results with unwanted attribute values.
    """

    def __init__(self, bad_values=[], attr='locator', exact_match=True):
        """
        :arg list bad_values: A list of values whose candidates we will
                              not accept results from (default [])
        :arg string attr: The attribute type on which to filter
        :arg bool exact_match: True if attribute must match a bad value exactly.
                               False if the bad value can be a substring of the
                               attribute value. In other words, if our Candidate
                               attribute is 'Postcode3' and one of the bad values
                               is 'Postcode' because we want something more precise,
                               like 'Address', we will not keep this candidate.
        """
        self._init_helper(vars())

    def process(self, candidates):
        if self.exact_match is True:
            return [c for c in candidates if getattr(c, self.attr) not in self.bad_values]
        else:
            return [c for c in candidates if not any(bv in getattr(c, self.attr) for bv in self.bad_values)]

    def __repr__(self):
        return '<%s: %s %s in %s>' % \
            (self.__class__.__name__, self.is_exact(), self.attr, self.bad_values)


class AttrListIncludes(_PostProcessor):
    """
    PostProcessor used to filter out results without desired attribute list items.

    Similar to `AttrFilter` but operates on attributes containing lists instead of scalar values.
    """

    def __init__(self, good_values=[], attr='entity_types'):
        """
        :arg list good_values: A list of values, one of which must be in the
                               attribute being filtered on (default [])
        :arg string attr: The attribute on which to filter
        """
        self._init_helper(vars())

    def process(self, candidates):
        return [c for c in candidates if any(gv in getattr(c, self.attr)
                                             for gv in self.good_values)]

    def __repr__(self):
        return '<%s: %s in %s>' % \
            (self.__class__.__name__, self.attr, self.good_values)


class AttrListExcludes(_PostProcessor):
    """
    PostProcessor used to ditch results with unwanted attribute list items.

    Similar to `AttrExclude` but operates on attributes containing lists instead of scalar values.
    """

    def __init__(self, bad_values=[], attr='entity_types'):
        """
        :arg list bad_values: A list of values, which cannot be in the
                              attribute being filtered on (default [])
        :arg string attr: The attribute on which to filter
        """
        self._init_helper(vars())

    def process(self, candidates):
        return [c for c in candidates if not any(bv in getattr(c, self.attr)
                                                 for bv in self.bad_values)]

    def __repr__(self):
        return '<%s: %s in %s>' % \
            (self.__class__.__name__, self.attr, self.bad_values)


class DupePicker(_PostProcessor):
    """
    PostProcessor used to choose the best candidate(s)
    where there are duplicates (or more than one result
    that is very similar*) among high-scoring candidates,
    such as an address.

    * When comparing attribute values, case and commas do not count.

    Usage Example:

    ================ ===== =======
    match_addr       score locator
    ================ ===== =======
    123 N Wood St    90    roof
    123 S Wood St    90    address
    123 N WOOD ST    77    address
    123, S Wood ST   85    roof
    ================ ===== =======

    Above, the first two results have the highest scores. We could just
    use those, because one of the two likely has the correct address.
    However, the second result does not have the most precise location
    for 123 S. Wood Street. While the fourth result does not score as
    high as the first two, its locator method is more desirable.
    Since the addresses are the same, we can assume that the fourth result
    will provide better data than the second.

    We can get a narrowed list as described above by running the process()
    method in the DupePicker() PostProcessor class as follows, assuming
    that the "candidates" is our list of candidates::

        dp = DupePicker(
            attr_dupes='match_addr',
            attr_sort='locator',
            ordered_list=['rooftop', 'address_point', 'address_range'])

        return dp.process(candidates)

    Output:

    ================ ===== =======
    match_addr       score locator
    ---------------- ----- -------
    123 N Wood St    90    roof
    123, S Wood ST   85    roof
    ================ ===== =======

    Output with return_clean=True:

    ================ ===== =======
    match_addr       score locator
    ---------------- ----- -------
    123 N WOOD ST    90    roof
    123 S WOOD ST    85    roof
    ================ ===== =======
    """

    def __init__(self, attr_dupes, attr_sort, ordered_list, return_clean=False):
        """
        :arg str attr_dupes: Property on which to look for duplicates.
        :arg str attr_sort: Property on which to sort using ordered_list
        :arg list ordered_list: A list of property values, from most desirable
                                to least desirable.
        :arg bool return_clean: Boolean indicating whether or not to
                                homogenize string values into uppercase
                                without commas.

        """
        self._init_helper(vars())

    def process(self, candidates):
        def cleanup(str_):
            """Returns string in uppercase and free of commas."""
            if type(str_) in (str, unicode):
                return str_.replace(',', '').upper()
            return str_

        # if there are no candidates, then there is nothing to do here
        if candidates == []:
            return []
        hi_score = ScoreSorter().process(candidates)[0].score
        hi_score_candidates = AttrFilter([hi_score], 'score').process(candidates)
        new_candidates = []
        for hsc in hi_score_candidates:
            # get candidates with same address, including the current one:
            attr_match = self.attr_dupes
            attr_match_test_val = cleanup(getattr(hsc, attr_match))
            # make a list of candidates that have essentially the same value for attr_match (like 123 Main & 123 MAIN)
            matching_candidates = [mc for mc in candidates if cleanup(getattr(mc, attr_match)) == attr_match_test_val]
            # sort them in the desired order so the first one has the best attribute value
            matching_candidates = AttrSorter(self.ordered_list, self.attr_sort).process(matching_candidates)
            # the best value available can be grabbed from the top result:
            best_attr_value = getattr(matching_candidates[0], attr_match)
            # now we can filter results that have best_attr_value:
            new_candidates_queue = AttrFilter([best_attr_value], attr_match).process(matching_candidates)
            # and append each one to our list of new candidates, if it's not there already:
            for nc in [nc for nc in new_candidates_queue if nc not in new_candidates]:
                if self.return_clean:
                    new_candidates.append(cleanup(nc))
                else:
                    new_candidates.append(nc)
        return new_candidates

    def __repr__(self):
        repr_ = '%s: SORT BY %s %s -> GROUP BY %s' % \
            (self.__class__.__name__, self.attr_sort, self.ordered_list, self.attr_dupes)
        if self.return_clean:
            repr_ += ' -> CLEAN'
        return '<%s>' % repr_


class GroupBy(_PostProcessor):
    """
    Groups results by a certain attribute by choosing the
    first occurrence in the list (this means you will want
    to sort ahead of time).


    attr   --  The attribute on which to combine results
               or a list or tuple of attributes where all
               attributes must match between candidates.
    """

    def __init__(self, attr='match_addr'):
        self._init_helper(vars())

    def process(self, candidates):
        if type(self.attr) in (tuple, list):
            return GroupByMultiple(attrs=self.attr).process(candidates)
        keepers = []
        for c_from_all in candidates[:]:
            matches = [c for c in candidates if getattr(c, self.attr) == getattr(c_from_all, self.attr)]
            if matches != []:
                keepers.append(matches[0])
                for m in matches:
                    candidates.remove(m)
        return keepers

    def __repr__(self):
        return '<%s: %s>' % \
            (self.__class__.__name__, self.attr)


class GroupByMultiple(_PostProcessor):
    """
    Groups results by a certain attribute by choosing the
    first occurrence in the list of candidates
    (this means you will want to sort ahead of time).


    attrs   --  A list or tuple of attributes on which to combine results
    """

    def __init__(self, attrs):
        self._init_helper(vars())

    def process(self, candidates):
        keepers = []
        for c_from_all in candidates[:]:
            matches = [c for c in candidates
                       if all([getattr(c, attr) == getattr(c_from_all, attr)
                               for attr in self.attrs])]
            if matches != []:
                keepers.append(matches[0])
                for m in matches:
                    candidates.remove(m)
        return keepers

    def __repr__(self):
        return '<%s: %s>' % \
            (self.__class__.__name__, self.attrs)


class SnapPoints(_PostProcessor):
    """
    Chooses the first of two or more points where they are within the given
    sphere-based great circle distance.
    """
    def __init__(self, distance=50):
        """
        :arg distance: maximum distance (in metres) between two points in which
                       the the first will be kept and the second thrown out
                       (default 50).
        """
        self.distance = distance

    def _get_distance(self, pnt1, pnt2):
        """Get distance in meters between two lat/long points"""
        lat1, lon1 = pnt1
        lat2, lon2 = pnt2
        radius = 6356752  # km
        dlat = math.radians(lat2 - lat1)
        dlon = math.radians(lon2 - lon1)
        a = math.sin(dlat / 2) * math.sin(dlat / 2) + math.cos(math.radians(lat1)) \
            * math.cos(math.radians(lat2)) * math.sin(dlon / 2) * math.sin(dlon / 2)
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        d = radius * c
        return d

    def _points_within_distance(self, pnt1, pnt2):
        """Returns true if lat/lon points are within given distance in metres."""
        if self._get_distance(pnt1, pnt2) <= self.distance:
            return True
        return False

    def process(self, candidates):
        keepers = []
        for c_from_all in candidates[:]:
            matches = [c for c in candidates if
                       self._points_within_distance((c_from_all.x, c_from_all.y), (c.x, c.y))]
            if matches != []:
                keepers.append(matches[0])
                for m in matches:
                    candidates.remove(m)
        return keepers

    def __repr__(self):
        return '<%s: distance=%sm>' % \
            (self.__class__.__name__, self.distance)



from datetime import datetime, timedelta
import logging

from omgeo.services.base import GeocodeService
from omgeo.places import Candidate
from omgeo.preprocessors import CancelIfPOBox
from omgeo.postprocessors import (AttrFilter, AttrRename, AttrSorter, UseHighScoreIfAtLeast,
                                  GroupBy, ScoreSorter)

logger = logging.getLogger(__name__)


class EsriWGS(GeocodeService):
    """
    Class to geocode using the `ESRI World Geocoding service
    <https://developers.arcgis.com/features/geocoding/>`_.

    This uses two endpoints -- one for single-line addresses,
    and one for multi-part addresses.

    An optional ``key`` parameter can be passed to the :class:`~omgeo.places.PlaceQuery`
    which will be passed as a ``magicKey`` to the find endpoint if
    using a single line address/text search. This allows omgeo
    to be used with the `Esri suggest endpoint
    <https://developers.arcgis.com/rest/geocode/api-reference/geocoding-suggest.htm>`_.

    .. warning::

        Based on tests using the magicKey parameter, it is
        recommended that a viewbox not be used with in conjuction
        with the magicKey. Additionally, address/search text passed
        via the query may be ignored when using a magicKey.

    An optional ``for_storage`` flag can be passed to the :class:`~omgeo.places.PlaceQuery` which
    will cause ``findStorage=true`` to be passed to the find endpoint.
    The Esri terms of service requires this if the returned point will be stored
    in a database.
    If you pass this flag, you `must` set the client_id and client_secret settings.

    Settings used by the EsriWGS GeocodeService object may include:
     * client_id --  The Client ID used to access Esri services.
     * client_secret --  The Client Secret used to access Esri services.

    """

    LOCATOR_MAP = {
        'PointAddress': 'rooftop',
        'StreetAddress': 'interpolation',
        'PostalExt': 'postal_specific',  # accept ZIP+4
        'Postal': 'postal'
    }

    DEFAULT_PREPROCESSORS = [CancelIfPOBox()]

    DEFAULT_POSTPROCESSORS = [
        AttrFilter(['PointAddress',
                    'StreetAddress',
                    # 'PostalExt',
                    # 'Postal'
                    ],
                   'locator_type'),
        # AttrExclude(['USA_Postal'], 'locator'), #accept postal from everywhere but US (need PostalExt)
        AttrSorter(['PointAddress',
                    'StreetAddress',
                    # 'PostalExt',
                    # 'Postal'
                    ],
                   'locator_type'),
        AttrRename('locator', LOCATOR_MAP),  # after filter to avoid searching things we toss out
        UseHighScoreIfAtLeast(99.8),
        ScoreSorter(),
        GroupBy('match_addr'),
        GroupBy(('x', 'y')),
    ]

    _endpoint = 'https://geocode.arcgis.com/arcgis/rest/services/World/GeocodeServer'

    def __init__(self, preprocessors=None, postprocessors=None, settings=None):
        preprocessors = EsriWGS.DEFAULT_PREPROCESSORS if preprocessors is None else preprocessors
        postprocessors = EsriWGS.DEFAULT_POSTPROCESSORS if postprocessors is None else postprocessors
        settings = {} if settings is None else settings

        if 'client_id' in settings and 'client_secret' in settings:
            self._authenticated = True
            self._client_id = settings['client_id']
            self._client_secret = settings['client_secret']
            self._token = None
        elif 'client_id' not in settings and 'client_secret' not in settings:
            self._authenticated = False
        else:
            raise Exception('Must specify both client_id and client_secret to use authentication')

        GeocodeService.__init__(self, preprocessors, postprocessors, settings)

    def _geocode(self, pq):
        """
        :arg PlaceQuery pq: PlaceQuery object to use for geocoding
        :returns: list of location Candidates
        """
        #: List of desired output fields
        #: See `ESRI docs <https://developers.arcgis.com/rest/geocode/api-reference/geocoding-geocode-addresses.htm>_` for details
        outFields = ('Loc_name',
                     # 'Shape',
                     'Score',
                     'Match_addr',  # based on address standards for the country
                     # 'Address', # returned by default
                     # 'Country' # 3-digit ISO 3166-1 code for a country. Example: Canada = "CAN"
                     # 'Admin',
                     # 'DepAdmin',
                     # 'SubAdmin',
                     # 'Locality',
                     # 'Postal',
                     # 'PostalExt',
                     'Addr_type',
                     # 'Type',
                     # 'Rank',
                     'AddNum',
                     'StPreDir',
                     'StPreType',
                     'StName',
                     'StType',
                     'StDir',
                     # 'Side',
                     # 'AddNumFrom',
                     # 'AddNumTo',
                     # 'AddBldg',
                     'City',
                     'Subregion',
                     'RegionAbbr',
                     'Postal',
                     'Country',
                     # 'Ymax',
                     # 'Ymin',
                     # 'Xmin',
                     # 'Xmax',
                     # 'X',
                     # 'Y',
                     'DisplayX',
                     'DisplayY',
                     # 'LangCode',
                     # 'Status',
                     )
        outFields = ','.join(outFields)
        query = dict(f='json',  # default HTML. Other options are JSON and KMZ.
                     outFields=outFields,
                     # outSR=WKID, defaults to 4326
                     maxLocations=20,  # default 1; max is 20
                     )

        # Postal-code only searches work in the single-line but not multipart geocoder
        # Remember that with the default postprocessors, postcode-level results will be eliminated
        if pq.query == pq.address == '' and pq.postal != '':
            pq.query = pq.postal

        if pq.query == '':  # multipart
            query = dict(query,
                         Address=pq.address,  # commonly represents the house number and street name of a complete address
                         Neighborhood=pq.neighborhood,
                         City=pq.city,
                         Subregion=pq.subregion,
                         Region=pq.state,
                         Postal=pq.postal,
                         # PostalExt=
                         CountryCode=pq.country,  # full country name or ISO 3166-1 2- or 3-digit country code
                         )
        else:  # single-line
            magic_key = pq.key if hasattr(pq, 'key') else ''
            query = dict(query,
                         singleLine=pq.query,  # This can be a street address, place name, postal code, or POI.
                         sourceCountry=pq.country,  # full country name or ISO 3166-1 2- or 3-digit country code
                         )
            if magic_key:
                query['magicKey'] = magic_key  # This is a lookup key returned from the suggest endpoint.

        if pq.bounded and pq.viewbox is not None:
            query = dict(query, searchExtent=pq.viewbox.to_esri_wgs_json())

        if self._authenticated:
            if self._token is None or self._token_expiration < datetime.utcnow():
                expiration = timedelta(hours=2)
                self._token = self.get_token(expiration)
                self._token_expiration = datetime.utcnow() + expiration

            query['token'] = self._token

        if getattr(pq, 'for_storage', False):
            query['forStorage'] = 'true'

        endpoint = self._endpoint + '/findAddressCandidates'
        response_obj = self._get_json_obj(endpoint, query)
        returned_candidates = []  # this will be the list returned
        try:
            locations = response_obj['candidates']
            for location in locations:
                c = Candidate()
                attributes = location['attributes']
                c.match_addr = attributes['Match_addr']
                c.locator = attributes['Loc_name']
                c.locator_type = attributes['Addr_type']
                c.score = attributes['Score']
                c.x = attributes['DisplayX']  # represents the actual location of the address.
                c.y = attributes['DisplayY']
                c.wkid = response_obj['spatialReference']['wkid']
                c.geoservice = self.__class__.__name__

                # Optional address component fields.
                for in_key, out_key in [('City', 'match_city'), ('Subregion', 'match_subregion'),
                                        ('RegionAbbr', 'match_region'), ('Postal', 'match_postal'),
                                        ('Country', 'match_country')]:
                    setattr(c, out_key, attributes.get(in_key, ''))
                setattr(c, 'match_streetaddr', self._street_addr_from_response(attributes))
                returned_candidates.append(c)
        except KeyError:
            pass
        return returned_candidates

    def _street_addr_from_response(self, attributes):
        """Construct a street address (no city, region, etc.) from a geocoder response.

        :param attributes: A dict of address attributes as returned by the Esri geocoder.
        """
        # The exact ordering of the address component fields that should be
        # used to reconstruct the full street address is not specified in the
        # Esri documentation, but the examples imply that it is this.
        ordered_fields = ['AddNum', 'StPreDir', 'StPreType', 'StName', 'StType', 'StDir']
        result = []
        for field in ordered_fields:
            result.append(attributes.get(field, ''))
        if any(result):
            return ' '.join([s for s in result if s])  # Filter out empty strings.
        else:
            return ''

    def get_token(self, expires=None):
        """
        :param expires: The time until the returned token expires.
            Must be an instance of :class:`datetime.timedelta`.
            If not specified, the token will expire in 2 hours.
        :returns: A token suitable for use with the Esri geocoding API
        """
        endpoint = 'https://www.arcgis.com/sharing/rest/oauth2/token/'
        query = {'client_id': self._client_id,
                 'client_secret': self._client_secret,
                 'grant_type': 'client_credentials'}

        if expires is not None:
            if not isinstance(expires, timedelta):
                raise Exception('If expires is provided it must be a timedelta instance')
            query['expiration'] = int(expires.total_seconds() / 60)

        response_obj = self._get_json_obj(endpoint, query, is_post=True)

        return response_obj['access_token']



from .base import GeocodeService
import logging
from omgeo.places import Candidate
from omgeo.preprocessors import ReplaceRangeWithNumber
from omgeo.postprocessors import (AttrFilter, AttrRename, AttrSorter, AttrMigrator,
                                  UseHighScoreIfAtLeast, GroupBy, ScoreSorter)

logger = logging.getLogger(__name__)


class Bing(GeocodeService):
    """
    Class to geocode using Bing services:
     * `Find a Location by Query <http://msdn.microsoft.com/en-us/library/ff701711.aspx>`_
     * `Find a Location by Address <http://msdn.microsoft.com/en-us/library/ff701714.aspx>`_

    Settings used by the Bing GeocodeService object may include:
     * api_key --  The API key used to access Bing services.

    """
    _endpoint = 'http://dev.virtualearth.net/REST/v1/Locations'

    DEFAULT_PREPROCESSORS = [
        ReplaceRangeWithNumber()
    ]

    DEFAULT_POSTPROCESSORS = [
        AttrMigrator('confidence', 'score',
                     {'High': 100, 'Medium': 85, 'Low': 50}),
        UseHighScoreIfAtLeast(100),
        AttrFilter(['Address', 'AdministrativeBuilding',
                    'AgriculturalStructure',
                    'BusinessName', 'BusinessStructure',
                    'BusStation', 'Camp', 'Church', 'CityHall',
                    'CommunityCenter', 'ConventionCenter',
                    'Courthouse', 'Factory', 'FerryTerminal',
                    'FishHatchery', 'Fort', 'Garden', 'Geyser',
                    'Heliport', 'IndustrialStructure',
                    'InformationCenter', 'Junction',
                    'LandmarkBuilding', 'Library', 'Lighthouse',
                    'Marina', 'MedicalStructure', 'MetroStation',
                    'Mine', 'Mission', 'Monument', 'Mosque',
                    'Museum', 'NauticalStructure', 'NavigationalStructure',
                    'OfficeBuilding', 'ParkAndRide', 'PlayingField',
                    'PoliceStation', 'PostOffice', 'PowerStation',
                    'Prison', 'RaceTrack', 'ReligiousStructure',
                    'RestArea', 'Ruin', 'ShoppingCenter', 'Site',
                    'SkiArea', 'Spring', 'Stadium', 'Temple',
                    'TouristStructure'], 'entity'),
        AttrRename('locator', dict(Rooftop='rooftop',
                                   Parcel='parcel',
                                   ParcelCentroid='parcel',
                                   Interpolation='interpolation',
                                   InterpolationOffset='interpolation_offset')),
        AttrSorter(['rooftop', 'parcel',
                    'interpolation_offset', 'interpolation'],
                   'locator'),
        AttrSorter(['Address'], 'entity'),
        ScoreSorter(),
        GroupBy(('x', 'y')),
        GroupBy('match_addr')]
    DEFAULT_POSTPROCESSORS = []

    def __init__(self, preprocessors=None, postprocessors=None, settings=None):
        preprocessors = Bing.DEFAULT_PREPROCESSORS if preprocessors is None else preprocessors
        postprocessors = Bing.DEFAULT_POSTPROCESSORS if postprocessors is None else postprocessors
        GeocodeService.__init__(self, preprocessors, postprocessors, settings)

    def _geocode(self, pq):
        if pq.query.strip() == '':
            # No single line query string; use address elements:
            query = {'addressLine': pq.address,
                     'locality': pq.city,
                     'adminDistrict': pq.state,
                     'postalCode': pq.postal,
                     'countryRegion': pq.country}
        else:
            query = {'query': pq.query}

        if pq.viewbox is not None:
            query = dict(query, **{'umv': pq.viewbox.to_bing_str()})
        if hasattr(pq, 'culture'):
            query = dict(query, c=pq.culture)
        if hasattr(pq, 'user_ip'):
            query = dict(query, uip=pq.user_ip)
        if hasattr(pq, 'user_lat') and hasattr(pq, 'user_lon'):
            query = dict(query, **{'ul': '%f,%f' % (pq.user_lat, pq.user_lon)})

        addl_settings = {'key': self._settings['api_key']}
        query = dict(query, **addl_settings)
        response_obj = self._get_json_obj(self._endpoint, query)
        returned_candidates = []  # this will be the list returned
        for r in response_obj['resourceSets'][0]['resources']:
            c = Candidate()
            c.entity = r['entityType']
            c.locator = r['geocodePoints'][0]['calculationMethod']  # ex. "Parcel"
            c.confidence = r['confidence']  # High|Medium|Low
            c.match_addr = r['name']  # ex. "1 Microsoft Way, Redmond, WA 98052"
            c.x = r['geocodePoints'][0]['coordinates'][1]  # long, ex. -122.13
            c.y = r['geocodePoints'][0]['coordinates'][0]  # lat, ex. 47.64
            c.wkid = 4326
            c.geoservice = self.__class__.__name__
            returned_candidates.append(c)
        return returned_candidates



import logging

from .base import GeocodeService
from omgeo.places import Candidate
from omgeo.preprocessors import ComposeSingleLine
from omgeo.postprocessors import AttrListExcludes

logger = logging.getLogger(__name__)


class Google(GeocodeService):
    """
    Class to geocode using Google's geocoding API.
    """
    _endpoint = 'https://maps.googleapis.com/maps/api/geocode/json'

    DEFAULT_PREPROCESSORS = [ComposeSingleLine()]

    DEFAULT_POSTPROCESSORS = [
        AttrListExcludes([
            # The Google docs note "establishment" as: typically indicates a place
            # that has not yet been categorized.
            #
            # In practice though, this seems to only appear when searching by a business / location name,
            # or a person associated with a business or location
            'establishment',
        ], 'entity_types')
    ]

    LOCATOR_MAPPING = {
        'ROOFTOP': 'rooftop',
        'RANGE_INTERPOLATED': 'interpolated',
    }

    def __init__(self, preprocessors=None, postprocessors=None, settings=None):
        preprocessors = self.DEFAULT_PREPROCESSORS if preprocessors is None else preprocessors
        postprocessors = self.DEFAULT_POSTPROCESSORS if postprocessors is None else postprocessors
        GeocodeService.__init__(self, preprocessors, postprocessors, settings)

    def _geocode(self, pq):
        params = {
            'address': pq.query,
            'key': self._settings['api_key']
        }

        if pq.country:
            params['components'] = 'country:' + pq.country
        if pq.viewbox:
            params['bounds'] = pq.viewbox.to_google_str()

        response_obj = self._get_json_obj(self._endpoint, params)
        return [self._make_candidate_from_result(r) for r in response_obj['results']]

    def _make_candidate_from_result(self, result):
        """ Make a Candidate from a Google geocoder results dictionary. """
        candidate = Candidate()
        candidate.match_addr = result['formatted_address']
        candidate.x = result['geometry']['location']['lng']
        candidate.y = result['geometry']['location']['lat']
        candidate.locator = self.LOCATOR_MAPPING.get(result['geometry']['location_type'], '')
        candidate.entity_types = result['types']
        candidate.partial_match = result.get('partial_match', False)

        component_lookups = {
            'city': {'type': 'locality', 'key': 'long_name'},
            'subregion': {'type': 'administrative_area_level_2', 'key': 'long_name'},
            'region': {'type': 'administrative_area_level_1', 'key': 'short_name'},
            'postal': {'type': 'postal_code', 'key': 'long_name'},
            'country': {'type': 'country', 'key': 'short_name'},
        }
        for (field, lookup) in component_lookups.items():
            setattr(candidate, 'match_' + field, self._get_component_from_result(result, lookup))
        candidate.geoservice = self.__class__.__name__
        return candidate

    def _get_component_from_result(self, result, lookup):
        """
        Helper function to get a particular address component from a Google result.

        Since the address components in results are an array of objects containing a types array,
        we have to search for a particular component rather than being able to look it up directly.

        Returns the first match, so this should be used for unique component types (e.g.
        'locality'), not for categories (e.g. 'political') that can describe multiple components.

        :arg dict result: A results dict with an 'address_components' key, as returned by the
                          Google geocoder.
        :arg dict lookup: The type (e.g. 'street_number') and key ('short_name' or 'long_name') of
                          the desired address component value.
        :returns: address component or empty string
        """
        for component in result['address_components']:
            if lookup['type'] in component['types']:
                return component.get(lookup['key'], '')
        return ''



from .base import GeocodeService
import json
import logging
from omgeo.places import Candidate

try:
    # python 3
    from urllib.parse import unquote
except ImportError:
    # python 2
    from urllib import unquote

logger = logging.getLogger(__name__)


class MapQuest(GeocodeService):
    """
    Class to geocode using MapQuest licensed services.
    """
    _endpoint = 'http://www.mapquestapi.com/geocoding/v1/address'

    def _geocode(self, pq):
        def get_appended_location(location, **kwargs):
            """Add key/value pair to given dict only if value is not empty string."""
            for kw in kwargs:
                if kwargs[kw] != '':
                    location = dict(location, **{kw: kwargs[kw]})
            return location
        location = {}
        location = get_appended_location(location, street=pq.query)
        if location == {}:
            location = get_appended_location(location, street=pq.address)
        location = get_appended_location(location, city=pq.city, county=pq.subregion,
                                         state=pq.state, postalCode=pq.postal, country=pq.country)
        json_ = dict(location=location)
        json_ = json.dumps(json_)
        logger.debug('MQ json: %s', json_)
        query = dict(key=unquote(self._settings['api_key']),
                     json=json_)
        if pq.viewbox is not None:
            query = dict(query, viewbox=pq.viewbox.to_mapquest_str())
        response_obj = self._get_json_obj(self._endpoint, query)
        logger.debug('MQ RESPONSE: %s', response_obj)
        returned_candidates = []  # this will be the list returned
        for r in response_obj['results'][0]['locations']:
            c = Candidate()
            c.locator = r['geocodeQuality']
            c.confidence = r['geocodeQualityCode']  # http://www.mapquestapi.com/geocoding/geocodequality.html
            match_addr_elements = ['street', 'adminArea5', 'adminArea3',
                                   'adminArea2', 'postalCode']  # similar to ESRI
            c.match_addr = ', '.join([r[k] for k in match_addr_elements if k in r])
            c.x = r['latLng']['lng']
            c.y = r['latLng']['lat']
            c.wkid = 4326
            c.geoservice = self.__class__.__name__
            returned_candidates.append(c)
        return returned_candidates


class MapQuestSSL(MapQuest):
    _endpoint = 'https://www.mapquestapi.com/geocoding/v1/address'



import copy
import logging
from omgeo.places import PlaceQuery
from omgeo.postprocessors import DupePicker, SnapPoints

logger = logging.getLogger(__name__)
stats_logger = logging.getLogger('omgeo.stats')

# Python 2/3 compatibility: set up 'unicode' for use in string type checking
try:
    unicode
except NameError:
    unicode = str


class Geocoder():
    """
    Class for building a custom geocoder using external APIs.
    """

    DEFAULT_SOURCES = [
        ['omgeo.services.EsriWGS', {}],
        ['omgeo.services.Nominatim', {}]
    ]
    DEFAULT_PREPROCESSORS = []
    DEFAULT_POSTPROCESSORS = [
        SnapPoints(),
        DupePicker('match_addr', 'locator',
                   ['rooftop', 'parcel', 'interpolation_offset', 'interpolation'])
    ]

    def _get_service_by_name(self, service_name):
        try:
            module, separator, class_name = service_name.rpartition('.')
            m = __import__(module)
            path = service_name.split('.')[1:]
            for p in path:
                m = getattr(m, p)
            return m
        except Exception as ex:
            raise Exception("%s" % (ex))

    def add_source(self, source):
        """
        Add a geocoding service to this instance.
        """
        geocode_service = self._get_service_by_name(source[0])
        self._sources.append(geocode_service(**source[1]))

    def remove_source(self, source):
        """
        Remove a geocoding service from this instance.
        """
        geocode_service = self._get_service_by_name(source[0])
        self._sources.remove(geocode_service(**source[1]))

    def set_sources(self, sources):
        """
        Creates GeocodeServiceConfigs from each str source
        """
        if len(sources) == 0:
            raise Exception('Must declare at least one source for a geocoder')
        self._sources = []
        for source in sources:  # iterate through a list of sources
            self.add_source(source)

    def __init__(self, sources=None, preprocessors=None, postprocessors=None,
                 waterfall=False):
        """
        :arg list sources: an array of GeocodeServiceConfig() parameters,
                           keyed by module name for the GeocodeService to use, e.g.::

                               [['esri_wgs', {}],
                                ['bing', {'settings': {'request_headers': {'User-Agent': 'Custom User Agent'} },
                                         'preprocessors': [],
                                         'postprocessors': []}],
                                 ...]

        :arg list preprocessors: list of universal preprocessors to use
        :arg list postprocessors: list of universal postprocessors to use
        :arg bool waterfall: sets default for waterfall on geocode() method (default ``False``)
        """

        self._preprocessors = Geocoder.DEFAULT_PREPROCESSORS \
            if preprocessors is None else preprocessors
        self._postprocessors = Geocoder.DEFAULT_POSTPROCESSORS \
            if postprocessors is None else postprocessors
        sources = Geocoder.DEFAULT_SOURCES if sources is None else sources
        self.set_sources(sources)
        self.waterfall = waterfall

    def geocode(self, pq, waterfall=None, force_stats_logging=False):
        """
        :arg PlaceQuery pq:  PlaceQuery object (required).
        :arg bool waterfall: Boolean set to True if all geocoders listed should
                             be used to find results, instead of stopping after
                             the first geocoding service with valid candidates
                             (defaults to self.waterfall).
        :arg bool force_stats_logging: Raise exception if stats logging fails (default False).
        :returns: Returns a dictionary including:
                   * candidates - list of Candidate objects
                   * upstream_response_info - list of UpstreamResponseInfo objects
        """

        waterfall = self.waterfall if waterfall is None else waterfall
        if type(pq) in (str, unicode):
            pq = PlaceQuery(pq)
        processed_pq = copy.copy(pq)

        for p in self._preprocessors:  # apply universal address preprocessing
            processed_pq = p.process(processed_pq)

        upstream_response_info_list = []
        processed_candidates = []
        for gs in self._sources:  # iterate through each GeocodeService
            candidates, upstream_response_info = gs.geocode(processed_pq)
            if upstream_response_info is not None:
                upstream_response_info_list.append(upstream_response_info)
            processed_candidates += candidates  # merge lists
            if waterfall is False and len(processed_candidates) > 0:
                break  # if >= 1 good candidate, don't go to next geocoder

        for p in self._postprocessors:  # apply univ. candidate postprocessing
            if processed_candidates == []:
                break  # avoid post-processing empty list
            processed_candidates = p.process(processed_candidates)

        result = dict(candidates=processed_candidates,
                      upstream_response_info=upstream_response_info_list)
        stats_dict = self.convert_geocode_result_to_nested_dicts(result)
        stats_dict = dict(stats_dict, original_pq=pq.__dict__)
        try:
            stats_logger.info(stats_dict)
        except Exception as exception:
            logger.error('Encountered exception while logging stats %s:\n%s', stats_dict, exception)
            if force_stats_logging:
                raise exception
        return result

    def get_candidates(self, pq, waterfall=None):
        """
        Geocode and return just the list of Candidate objects.
        """
        return self.geocode(pq, waterfall)['candidates']

    def convert_geocode_result_to_nested_dicts(self, result):
        def get_uri_dict(uri_item):
            uri_dict = copy.copy(uri_item).__dict__
            uri_dict['processed_pq'] = uri_dict['processed_pq'].__dict__
            return uri_dict
        uri_set = [get_uri_dict(uri_item) for uri_item in result['upstream_response_info']]
        return dict(candidates=[candidate.__dict__ for candidate in result['candidates']],
                    upstream_response_info=uri_set)



class Viewbox():
    """
    Class representing a bounding box.
    Defaults to maximum bounds for WKID 4326.
    """
    def convert_srs(self, new_wkid):
        """Return a new Viewbox object with the specified SRS."""
        return self  # not yet implemented

    def __init__(self, left=-180, top=90, right=180, bottom=-90, wkid=4326):
        """
        :arg left: Minimum X value (default ``-180``)
        :arg top: Maximum Y value (default ``90``)
        :arg right: Maximum X value (default ``180``)
        :arg bottom: Minimum Y value (default ``-90``)
        :arg wkid: Well-known ID for spatial reference system (default ``4326``)
        """
        bounds = left, right, bottom, top
        if not all([isinstance(x, (int, float)) for x in bounds]):
            raise ValueError('One or more bounds (%s) is not a real number.' % bounds)
        if left > right:
            raise ValueError('Left x-coord must be less than right x-coord.')
        if bottom > top:
            raise ValueError('Bottom y-coord must be less than top y-coord.')
        for k in list(locals().keys()):
            if k != 'self':
                setattr(self, k, locals()[k])

    def to_bing_str(self):
        """
        Convert Viewbox object to a string that can be used by Bing
        as a query parameter.
        """
        vb = self.convert_srs(4326)
        return '%s,%s,%s,%s' % (vb.bottom, vb.left, vb.top, vb.right)

    def to_pelias_dict(self):
        """
        Convert Viewbox object to a string that can be used by Pelias
        as a query parameter.
        """
        vb = self.convert_srs(4326)
        return {
            'boundary.rect.min_lat': vb.bottom,
            'boundary.rect.min_lon': vb.left,
            'boundary.rect.max_lat': vb.top,
            'boundary.rect.max_lon': vb.right
        }

    def to_google_str(self):
        """ Convert to Google's bounds format: 'latMin,lonMin|latMax,lonMax' """
        vb = self.convert_srs(4326)
        return '%s,%s|%s,%s' % (vb.bottom, vb.left, vb.top, vb.right)

    def to_mapquest_str(self):
        """
        Convert Viewbox object to a string that can be used by
        `MapQuest <http://www.mapquestapi.com/geocoding/#options>`_
        as a query parameter.
        """
        vb = self.convert_srs(4326)
        return '%s,%s,%s,%s' % (vb.left, vb.top, vb.right, vb.bottom)

    def to_esri_wgs_json(self):
        """
        Convert Viewbox object to a JSON string that can be used
        by the ESRI World Geocoding Service as a parameter.
        """
        try:
            return ('{ "xmin" : %s, '
                    '"ymin" : %s, '
                    '"xmax" : %s, '
                    '"ymax" : %s, '
                    '"spatialReference" : {"wkid" : %d} }'
                    % (self.left,
                       self.bottom,
                       self.right,
                       self.top,
                       self.wkid))
        except ValueError:
            raise Exception('One or more values could not be cast to a number. '
                            'Four bounding points must be real numbers. '
                            'WKID must be an integer.')

    def __repr__(self):
        top = "y=%s" % self.top
        right = "x=%s" % self.right
        bottom = "y=%s" % self.bottom
        left = "x=%s" % self.left

        def lbl(str_, align='L'):
            MAX_CHARS = 8
            str_len = len(str_)
            if str_len > MAX_CHARS:
                return str_[:MAX_CHARS]
            if align == 'L':
                return str_
            num_spaces = (MAX_CHARS - str_len)  # num spaces to pad right-aligned
            if align == 'C':
                num_spaces = int(num_spaces / 2)
            padding = num_spaces * ' '
            return '%s%s' % (padding, str_)

        return '          %s\n'\
               '        ------------\n'\
               '        |          |\n'\
               '%s|          |%s\n'\
               '        |          |\n'\
               '        ------------\n'\
               '          %s' % (lbl(top, 'C'), lbl(left, 'R'), lbl(right, 'L'), lbl(bottom, 'C'))


class PlaceQuery():
    """
    Class representing an address or place that will be passed to geocoders.
    """
    def __init__(self, query='', address='', neighborhood='', city='',
                 subregion='', state='', postal='', country='',
                 viewbox=None, bounded=False, **kwargs):
        """
        :arg str query: A string containing the query to parse
                        and match to a coordinate on the map.
                        *ex: "340 N 12th St Philadelphia PA 19107"
                        or "Wolf Building, Philadelphia"*
        :arg str address: A string for the street line of an address.
                          *ex: "340 N 12th St"*
        :arg str neighborhood: A string for the subdivision of a city. Not used
                               in US addresses, but used in Mexico and other places.
        :arg str city: A string specifying the populated place for the address.
                       This commonly refers to a city, but may refer to a suburb
                       or neighborhood in certain countries.
        :arg str subregion: A string for a region between the city and state level.
                            Not used in US addresses.
        :arg str state: A string for the state, province, territory, etc.
        :arg str postal: A string for the postal / ZIP Code
        :arg str country: A string for the country or region. Because the geocoder
                          uses the country to determine which geocoding service to use,
                          this is strongly recommended for efficency. ISO alpha-2 is
                          preferred, and is required by some geocoder services.
        :arg Viewbox viewbox: A Viewbox object indicating the preferred area
                              to find search results (default ``None``)
        :arg bool bounded: Boolean indicating whether or not to only
                           return candidates within the given Viewbox (default ``False``)

        :key float user_lat: A float representing the Latitude of the end-user.
        :key float user_lon: A float representing the Longitude of the end-user.
        :key str user_ip: A string representing the IP address of the end-user.
        :key str culture: Culture code to be used for the request (used by Bing).
                          For example, if set to 'de', the country for a U.S. address
                          would be returned as "Vereinigte Staaten Von Amerika"
                          instead of "United States".
    """
        for k in list(locals().keys()):
            if k not in ['self', 'kwargs']:
                setattr(self, k, locals()[k])
        if query == '' and address == '' and city == '' and state == '' and postal == '':
            raise Exception('Must provide query or one or more of address, city, state, and postal.')
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __repr__(self):
        return '<%s%s %s>' % (self.query, self.address, self.postal)


class Candidate():
    """
    Class representing a candidate address returned from geocoders.
    Accepts arguments defined below, plus informal keyword arguments.

    Usage Example::

        c = Candidate('US_RoofTop', 91.5, '340 N 12th St, Philadelphia, PA, 19107',
            '-75.16', '39.95', some_key_foo='bar')
    """

    def __init__(self, locator='', score=0, match_addr='', x=None, y=None,
                 wkid=4326, **kwargs):
        """
        :arg locator: Locator used for geocoding (default ``''``)

                      We try to standardize this to
                       * ``rooftop``,
                       * ``interpolated``,
                       * ``postal_specific``, and
                       * ``postal``.

        :arg score: Standardized score (default ``0``)
        :arg str match_addr: Address returned by geocoder (default ``''``)
        :arg x: X-coordinate (longitude for lat-lon SRS) (default ``None``)
        :arg y: Y-coordinate (latitude for lat-lon SRS) (default ``None``)
        :arg wkid: Well-known ID for spatial reference system (default ``4326``)

        Keyword arguments can be added in order to be able to use postprocessors
        with API output fields are not well-fitted for one of the definitions
        above.

        If possible, it is suggested for geocoders to additionally return the following
        address components:
            * match_streetaddr (the street address, e.g. '340 N 12th Street')
            * match_city
            * match_subregion (county)
            * match_region (state / province)
            * match_postal
            * match_country
        However, these are not required. Currently the EsriWGS and US Census geocoders
        return these values.
        """

        for k in list(locals().keys()):
            if k not in ['self', 'kwargs']:
                setattr(self, k, locals()[k])
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __repr__(self):
        if self.match_addr == '':
            match_addr = '(no address specified)'
        else:
            match_addr = self.match_addr

        if self.x is None:
            x = '(no x coord specified)'
        else:
            x = self.x

        if self.y is None:
            y = '(no y coord specified)'
        else:
            y = self.y

        geoservice = '%s' % getattr(self, 'geoservice', '(no geoservice specified')
        return '<%s (%s, %s) %s>' % (match_addr, x, y, geoservice)



import re

from .base import GeocodeService
import logging
from omgeo.places import Candidate

logger = logging.getLogger(__name__)


class USCensus(GeocodeService):

    # set endpoint based on whether we geocode by single-line address, or with keyed components
    _endpoint = ''
    _endpoint_base = 'https://geocoding.geo.census.gov/geocoder/locations/'

    def _geocode(self, pq):
        query = {
            'format': 'json',
            'benchmark': 'Public_AR_Current'
        }

        if pq.query:
            _this_endpoint = '%s%s' % (self._endpoint_base, 'onelineaddress')
            query['address'] = pq.query
        else:
            _this_endpoint = '%s%s' % (self._endpoint_base, 'address')
            query['street'] = pq.address
            query['city'] = pq.city
            query['state'] = pq.state
            query['zip'] = pq.postal

        logger.debug('CENSUS QUERY: %s', query)
        response_obj = self._get_json_obj(_this_endpoint, query)
        logger.debug('CENSUS RESPONSE: %s', response_obj)

        returned_candidates = []  # this will be the list returned
        for r in response_obj['result']['addressMatches']:
            c = Candidate()
            c.match_addr = r['matchedAddress']
            c.x = r['coordinates']['x']
            c.y = r['coordinates']['y']
            c.geoservice = self.__class__.__name__
            # Optional address component fields.
            for in_key, out_key in [('city', 'match_city'), ('state', 'match_region'),
                                    ('zip', 'match_postal')]:
                setattr(c, out_key, r['addressComponents'].get(in_key, ''))
            setattr(c, 'match_subregion', '')  # No county from Census geocoder.
            setattr(c, 'match_country', 'USA')  # Only US results from Census geocoder
            setattr(c, 'match_streetaddr', self._street_addr_from_response(r))
            returned_candidates.append(c)
        return returned_candidates

    def _street_addr_from_response(self, match):
        """Construct a street address (no city, region, etc.) from a geocoder response.

        :param match: The match object returned by the geocoder.
        """
        # Same caveat as above regarding the ordering of these fields; the
        # documentation is not explicit about the correct ordering for
        # reconstructing a full address, but implies that this is the ordering.
        ordered_fields = ['preQualifier', 'preDirection', 'preType', 'streetName',
                          'suffixType', 'suffixDirection', 'suffixQualifier']
        result = []
        # The address components only contain a from and to address, not the
        # actual number of the address that was matched, so we need to cheat a
        # bit and extract it from the full address string. This is likely to
        # miss some edge cases (hopefully only a few since this is a US-only
        # geocoder).
        addr_num_re = re.match(r'([0-9]+)', match['matchedAddress'])
        if not addr_num_re:  # Give up
            return ''
        result.append(addr_num_re.group(0))
        for field in ordered_fields:
            result.append(match['addressComponents'].get(field, ''))
        if any(result):
            return ' '.join([s for s in result if s])  # Filter out empty strings.
        else:
            return ''



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os
import sys
import unittest
from omgeo import Geocoder
from omgeo.places import Viewbox, PlaceQuery, Candidate
from omgeo.preprocessors import (CancelIfPOBox, CancelIfRegexInAttr, CountryPreProcessor,
                                 RequireCountry, ParseSingleLine, ReplaceRangeWithNumber)
from omgeo.postprocessors import (AttrFilter, AttrExclude, AttrRename,
                                  AttrSorter, AttrReverseSorter, UseHighScoreIfAtLeast,
                                  GroupBy, ScoreSorter, SnapPoints, AttrListIncludes, AttrListExcludes)

BING_MAPS_API_KEY = os.getenv("BING_MAPS_API_KEY")
MAPQUEST_API_KEY = os.getenv("MAPQUEST_API_KEY")
PELIAS_API_KEY = os.getenv("PELIAS_API_KEY")
GOOGLE_API_KEY = os.getenv("GOOGLE_API_KEY")
ESRI_CLIENT_ID = os.getenv("ESRI_CLIENT_ID")
ESRI_CLIENT_SECRET = os.getenv("ESRI_CLIENT_SECRET")

logger = logging.getLogger(__name__)
logging.basicConfig(level='ERROR')


class OmgeoTestCase(unittest.TestCase):
    def assertEqual_(self, output, expected):
        """assertEqual with built-in error message"""
        self.assertEqual(output, expected, 'Expected "%s". Got "%s".' % (expected, output))

    def assertEqualCI_(self, output, expected, strip_commas=False):
        """Case-insensitive assertEqual with built-in error message"""
        self.assertEqual_(str(output).upper(), str(expected).upper())

    def assertOneCandidate(self, candidates):
        count = len(candidates)
        self.assertEqual(count > 0, True, 'No candidates returned.')
        self.assertEqual(count > 1, False, 'More than one candidate returned.')


class GeocoderTest(OmgeoTestCase):
    """Tests using various geocoding APIs. Requires internet connection."""
    g = None  # not set until set up
    BING_KEY_REQUIRED_MSG = 'Enter a Bing Maps API key to run the Bing tests.'
    MAPQUEST_KEY_REQUIRED_MSG = 'Enter a MapQuest API key to run the MapQuest tests. '\
                                'Keys can be obtained at http://developer.mapquest.com/.'
    PELIAS_KEY_REQUIRED_MSG = 'Enter a Pelias service search API key to run Pelias tests. '\
                              'Keys can be obtained at https://geocode.earth/.'
    GOOGLE_KEY_REQUIRED_MSG = 'Enter a Google API key to run Google tests.'
    ESRI_KEY_REQUIRED_MSG = 'Enter a Esri Client ID & Secret to run authenticated Esri tests.'

    def setUp(self):
        # Viewbox objects - callowhill is from BSS Spring Garden station to Wash. Sq.
        vb = {'callowhill': Viewbox(-75.162628, 39.962769, -75.150963, 39.956322)}
        # PlaceQuery objects
        self.pq = {  # North American Addresses:
            'azavea': PlaceQuery('340 N 12th St Ste 402 Philadelphia PA'),
            'ambiguous_azavea': PlaceQuery('340 12th St Ste 402 Philadelphia PA'),
            'zip_plus_4_in_postal_plus_country': PlaceQuery(postal='19127-1115', country='US'),
            'wolf': PlaceQuery('Wolf Building'),
            'wolf_philly': PlaceQuery('Wolf Building, Philadelphia PA'),
            'wolf_bounded': PlaceQuery('Wolf Building', bounded=True, viewbox=vb['callowhill']),
            'bounded_340_12th': PlaceQuery('340 12th St, Philadelphia PA',
                                           bounded=True, viewbox=vb['callowhill']),
            'alpha_774R_W_Central_Ave': PlaceQuery('774R W Central Ave Alpha NJ'),
            'alpha_774_W_Central_Ave_Rear': PlaceQuery('774 W Central Ave Rear, Alpha NJ'),
            '8_kirkbride': PlaceQuery('8 Kirkbride Rd 08822'),
            'george_washington': PlaceQuery('201 W Montmorency Blvd, George, Washington'),
            'pine_needles_dr': PlaceQuery('11761 pine needles providence forge'),
            'pine_needles_ct': PlaceQuery('5328 pine needles providence forge'),
            'pine_needles_terr': PlaceQuery('5359 pine needles providence forge'),
            'moorestown_hyphenated': PlaceQuery('111-113 W Main St Moorestown NJ'),
            'willow_street': PlaceQuery('2819F Willow Street Pike Willow Street PA'),
            'willow_street_parts': PlaceQuery(address='2819F Willow Street Pike',
                                              city='Willow Street', state='PA', country='US'),
            'quebec': PlaceQuery('756 Rue Berri Montreal QC', country='CA'),
            'quebec_accent': PlaceQuery('527 Ch. Beauséjour, Saint-Elzéar-de-Témiscouata QC'),
            'quebec_hyphenated': PlaceQuery('227-227A Rue Commerciale, Saint-Louis-du-Ha! Ha! QC'),
            'senado_mx': PlaceQuery('Paseo de la Reforma 135, Tabacalera, Cuauhtémoc, Distrito Federal, 06030'),
            'senado_mx_struct': PlaceQuery(address='Paseo de la Reforma 135', neighborhood='Tabacalera, Cuauhtémoc', subregion='', state='Distrito Federal', postal='06030', country='MX'),
            'robert_cheetham': PlaceQuery('Robert Cheetham, Philadelphia'),
            # European Addresses:
            'london_pieces': PlaceQuery(address='31 Maiden Lane', city='London', country='UK'),
            'london_one_line': PlaceQuery('31 Maiden Lane, London WC2E', country='UK'),
            'london_pieces_hyphenated': PlaceQuery(address='31-32 Maiden Lane', city='London',
                                                   country='UK'),
            'london_one_line_hyphenated': PlaceQuery('31-32 Maiden Lane London WC2E', country='UK'),
            # Oceanian Addresses:
            'karori': PlaceQuery('102 Karori Road Karori Wellington', country='NZ'),
        }

        if BING_MAPS_API_KEY is not None:
            bing_settings = dict(api_key=BING_MAPS_API_KEY)
            self.g_bing = Geocoder([['omgeo.services.Bing', {'settings': bing_settings}]])

        if MAPQUEST_API_KEY is not None:
            mapquest_settings = dict(api_key=MAPQUEST_API_KEY)
            self.g_mapquest = Geocoder([['omgeo.services.MapQuest',
                                       {'settings': mapquest_settings}]])
            self.g_mapquest_ssl = Geocoder([['omgeo.services.MapQuestSSL',
                                           {'settings': mapquest_settings}]])

        if PELIAS_API_KEY is not None:
            pelias_settings = dict(api_key=PELIAS_API_KEY)
            self.g_pelias = Geocoder([['omgeo.services.Pelias', {'settings': pelias_settings}]])

        if GOOGLE_API_KEY is not None:
            self.g_google = Geocoder([['omgeo.services.Google',
                                     {'settings': {'api_key': GOOGLE_API_KEY}}]])
            self.g_google_wo_postprocess = Geocoder(
                [['omgeo.services.Google',
                  {'settings': {'api_key': GOOGLE_API_KEY}, 'postprocessors': []}]]
            )

        #: main geocoder used for tests, using default APIs
        self.g = Geocoder()

        # geocoders using individual services
        self.g_esri_wgs = Geocoder([['omgeo.services.EsriWGS', {}]])

        if ESRI_CLIENT_ID is not None and ESRI_CLIENT_SECRET is not None:
            self.g_esri_wgs_auth = Geocoder([['omgeo.services.EsriWGS',
                                              {'settings': {
                                                  'client_id': ESRI_CLIENT_ID,
                                                  'client_secret': ESRI_CLIENT_SECRET
                                              }}]])

        if MAPQUEST_API_KEY is not None:  # MapQuest's open Nominatime API now also requires a key
            self.g_nom = Geocoder([['omgeo.services.Nominatim', {}]])

        self.g_census = Geocoder([['omgeo.services.USCensus', {'settings': {'timeout': 30}}]])

        ESRI_WGS_LOCATOR_MAP = {'PointAddress': 'rooftop',
                                'StreetAddress': 'interpolation',
                                'PostalExt': 'postal_specific',  # accept ZIP+4
                                'Postal': 'postal'}
        ESRI_WGS_POSTPROCESSORS_POSTAL_OK = [
            AttrExclude(['USA.Postal'], 'locator'),  # accept postal from everywhere but US (need PostalExt)
            AttrFilter(['PointAddress', 'StreetAddress', 'PostalExt', 'Postal'], 'locator_type'),
            AttrSorter(['PointAddress', 'StreetAddress', 'PostalExt', 'Postal'], 'locator_type'),
            AttrRename('locator', ESRI_WGS_LOCATOR_MAP),  # after filter to avoid searching things we toss out
            UseHighScoreIfAtLeast(99.8),
            ScoreSorter(),
            GroupBy('match_addr'),
            GroupBy(('x', 'y')),
        ]
        GEOCODERS_POSTAL_OK = [['omgeo.services.EsriWGS', {'postprocessors': ESRI_WGS_POSTPROCESSORS_POSTAL_OK}]]
        self.g_esri_wgs_postal_ok = Geocoder(GEOCODERS_POSTAL_OK)

        #: geocoder with fast timeout
        self.impatient_geocoder = Geocoder(
            [['omgeo.services.EsriWGS', {'settings': {'timeout': 0.001}}]])

    def tearDown(self):
        pass

    def test_geocode_azavea(self):
        candidates = self.g.get_candidates(self.pq['azavea'])
        self.assertOneCandidate(candidates)

    def test_impatiently_geocode_azavea(self):
        candidates = self.impatient_geocoder.get_candidates(self.pq['azavea'])
        self.assertEqual(len(candidates) == 0, True,
                         'Candidates were unexpectedly returned in under 10ms.')

    @unittest.skipIf(BING_MAPS_API_KEY is None, BING_KEY_REQUIRED_MSG)
    def test_geocode_snap_points_2(self):
        """
        Bing geocoder expected to return the same place twice -- one with city as Alpha,
        and one with city as Phillipsburg. This test checks that only one is picked.
        """
        candidates = self.g_bing.get_candidates(self.pq['alpha_774_W_Central_Ave_Rear'])
        self.assertOneCandidate(candidates)

    def test_geocode_esri_wgs_senado_mx(self):
        """
        Attempt to geocode ``Paseo de la Reforma 135, Tabacalera,
        Cuauhtémoc, Distrito Federal, 06030``.
        """
        candidates = self.g_esri_wgs.get_candidates(self.pq['senado_mx'])
        self.assertOneCandidate(candidates)
        search_text = 'Paseo de la Reforma 135'
        self.assertEqual(search_text in candidates[0].match_addr, True,
                         '"%s" not found in match_addr. Got "%s".'
                         % (search_text, candidates[0].match_addr))

    def test_geocode_structured_esri_wgs_senado_mx(self):
        """
        Attempt to geocode ``Paseo de la Reforma 135, Tabacalera,
        Cuauhtémoc, Distrito Federal, 06030`` using a structured query to
        EsriWGS.
        """
        candidates = self.g_esri_wgs.get_candidates(self.pq['senado_mx_struct'])
        self.assertOneCandidate(candidates)
        search_text = 'Paseo de la Reforma 135'
        self.assertEqual(search_text in candidates[0].match_addr, True,
                         '"%s" not found in match_addr. Got "%s".'
                         % (search_text, candidates[0].match_addr))

    def test_geocode_esri_wgs_340_12th_bounded(self):
        """
        Trying to geocode ``340 12th St, Philadelphia PA`` would normally return results
        for both ``340 N 12th St`` and ``340 S 12th St``. Using a bounding box around Callowhill,
        we should only get the former.
        """
        candidates = self.g_esri_wgs.get_candidates(self.pq['bounded_340_12th'])
        self.assertOneCandidate(candidates)
        self.assertEqual('340 N 12th' in candidates[0].match_addr, True,
                         '"340 N 12th" not found in match_addr. Got "%s"' % candidates[0].match_addr)

    def test_geocode_esri_wgs_zip_plus_4(self):
        """Check that geocoding 19127-1112 returns one result."""
        candidates = self.g_esri_wgs_postal_ok.get_candidates(self.pq['zip_plus_4_in_postal_plus_country'])
        self.assertOneCandidate(candidates)

    def test_geocode_esri_wgs_multipart(self):
        """Check that geocoding multipart address returns one result."""
        candidates = self.g_esri_wgs.get_candidates(self.pq['willow_street_parts'])
        self.assertOneCandidate(candidates)

    @unittest.skipIf(ESRI_CLIENT_SECRET is None or ESRI_CLIENT_ID is None, ESRI_KEY_REQUIRED_MSG)
    def test_geocode_esri_wgs_auth(self):
        """Test that using authentication with the ESRI WGS geocoder is working"""
        candidates = self.g_esri_wgs_auth.get_candidates(self.pq['azavea'])
        self.assertOneCandidate(candidates)

    def test_esri_short_region(self):
        """Ensure that Esri uses region abbreviations"""
        candidate = self.g_esri_wgs.get_candidates(self.pq["azavea"])[0]
        self.assertEqual(candidate.match_region, "PA")

    def test_google_short_region(self):
        """Ensure that Google uses region abbreviations"""
        candidate = self.g_google.get_candidates(self.pq["azavea"])[0]
        self.assertEqual(candidate.match_region, "PA")

    @unittest.skipIf(BING_MAPS_API_KEY is None, BING_KEY_REQUIRED_MSG)
    def test_geocode_bing(self):
        """Test Azavea's address using Bing geocoder"""
        candidates = self.g_bing.get_candidates(self.pq['azavea'])
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    @unittest.skip("FIXME")
    def test_geocode_mapquest(self):
        """Test Azavea's address using MapQuest geocoder."""
        candidates = self.g_mapquest.get_candidates(self.pq['azavea'])
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    @unittest.skip("FIXME")
    def test_geocode_mapquest_ssl(self):
        """Test Azavea's address using secure MapQuest geocoder."""
        candidates = self.g_mapquest_ssl.get_candidates(self.pq['azavea'])
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    @unittest.skipIf(PELIAS_API_KEY is None, PELIAS_KEY_REQUIRED_MSG)
    def test_geocode_pelias(self):
        """Test Azavea's address using Pelias geocoder"""
        candidates = self.g_pelias.get_candidates(self.pq['azavea'])
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    @unittest.skip("FIXME")
    def test_geocode_nom(self):
        """
        Test 1200 Callowhill Street using Nominatim geocoder.
        Also check to make sure coordinate values are floats and not some other data type.
        """
        candidates = self.g_nom.get_candidates(PlaceQuery('1200 Callowhill St, Philadelphia, PA'))
        x_type = type(candidates[0].x)
        y_type = type(candidates[0].y)
        self.assertEqual(x_type == float, True, 'x coord is of type %s instead of float' % x_type)
        self.assertEqual(y_type == float, True, 'y coord is of type %s instead of float' % y_type)
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    def test_geocode_census(self):
        """Test Azavea's address using US Census geocoder."""
        candidates = self.g_census.get_candidates(PlaceQuery('1200 Callowhill St, Philadelphia, PA'))
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    def test_EsriWGS_address_components(self):
        """Make sure EsriWGS returns address components"""
        candidate = self.g_esri_wgs.get_candidates(self.pq['azavea'])[0]
        self._test_address_components(candidate)

    def test_census_address_components(self):
        """Make sure census geocoder returns address components"""
        candidate = self.g_census.get_candidates(self.pq['azavea'])[0]
        self._test_address_components(candidate)

    def test_geocode_dupepicker(self):
        """
        Check that '340 12th St returns results'
        """
        candidates = self.g.get_candidates(self.pq['ambiguous_azavea'])
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    @unittest.skipIf(BING_MAPS_API_KEY is None, BING_KEY_REQUIRED_MSG)
    def test_geocode_karori(self):
        """
        Check that '102 Karori Road Karori Wellington' returns an address
        with the correct house number and postcode.
        """
        candidates = self.g_bing.get_candidates(self.pq['karori'])
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')
        self.assertEqual(
            any([('102' in c.match_addr and '6012' in c.match_addr) for c in candidates]),
            True, 'Could not find bldg. no. "102" and postcode "6012" in any address.')

    def _test_address_components(self, candidate):
        for field in ['match_streetaddr', 'match_city', 'match_subregion', 'match_region',
                      'match_postal', 'match_country']:
            self.assertIsNotNone(getattr(candidate, field, None),
                                 msg='Missing address component %s' % field)

    def _test_geocode_results_all_(self, verbosity=0, geocoder=Geocoder(),
                                   expected_results=16):
        """
        Geocode a list of addresses. Some of these only work with Bing so
        fewer results are expected when Bing is not used as a geocoder.
        """
        if verbosity > 1:
            logger.setLevel(logging.INFO)

        queries_with_results = 0
        for place in self.pq:
            logger.info(place)
            logger.info(len(place) * '-')
            candidates = geocoder.get_candidates(self.pq[place])
            if len(candidates) == 0:
                logger.info('Input: %s\n(no results)' % self.pq[place].query)
            else:
                queries_with_results += 1
                logger.info('Input:  %s' % self.pq[place].query)
                logger.info([
                    'Output: %r (%s %s)\n' % (
                        c.match_addr,
                        c.geoservice,
                        [c.locator, c.score, c.confidence, c.entity])
                    for c in candidates])
        self.assertEqual(expected_results, queries_with_results,
                         'Got results for %d of %d queries.' % (queries_with_results, len(self.pq)))

    def _test_geocode_results_all(self):
        if BING_MAPS_API_KEY is None:
            expected_results = 16
        else:
            self.g.add_source(['omgeo.services.Bing', {'settings': {'api_key': BING_MAPS_API_KEY}}])
            expected_results = len(self.pq)
        self._test_geocode_results_all_(geocoder=self.g, expected_results=expected_results)

    @unittest.skipIf(GOOGLE_API_KEY is None, GOOGLE_KEY_REQUIRED_MSG)
    def test_google_geocode_azavea(self):
        candidates = self.g_google.get_candidates(self.pq['azavea'])
        self.assertOneCandidate(candidates)

    @unittest.skipIf(GOOGLE_API_KEY is None, GOOGLE_KEY_REQUIRED_MSG)
    def test_google_geocode_multipart(self):
        """Check that geocoding multipart address returns one result."""
        candidates = self.g_google.get_candidates(self.pq['willow_street_parts'])
        self.assertOneCandidate(candidates)

    @unittest.skipIf(GOOGLE_API_KEY is None, GOOGLE_KEY_REQUIRED_MSG)
    def test_google_country_filter(self):
        candidates = self.g_google.get_candidates(PlaceQuery('York', country='US'))
        self.assertOneCandidate(candidates)
        self.assertEqual(candidates[0].match_region, 'PA')
        candidates = self.g_google_wo_postprocess.get_candidates(PlaceQuery('York', country='GB'))
        self.assertOneCandidate(candidates)
        self.assertEqual(candidates[0].match_country, 'GB')

    @unittest.skipIf(GOOGLE_API_KEY is None, GOOGLE_KEY_REQUIRED_MSG)
    def test_google_geocode_without_postprocessor_allows_people(self):
        candidates = self.g_google_wo_postprocess.get_candidates(self.pq['robert_cheetham'])
        self.assertEqual(len(candidates) > 0, True, 'No candidates returned.')

    @unittest.skipIf(GOOGLE_API_KEY is None, GOOGLE_KEY_REQUIRED_MSG)
    def test_google_geocode_excludes_people(self):
        candidates = self.g_google.get_candidates(self.pq['robert_cheetham'])
        self.assertEqual(len(candidates) == 0, True, 'Candidate(s) unexpectedly returned.')


class GeocoderProcessorTest(OmgeoTestCase):
    """Tests using various pre- and post-processors."""
    def setUp(self):
        # places
        self.pq_us = PlaceQuery('1200 Callowhill St, Philadelphia, PA 19107')
        self.pq_uk = PlaceQuery('32 Bond Road, Ste A, Surbiton, Surrey KT6')
        self.pq_uk_with_country_UK = PlaceQuery('32 Bond Road, Ste A, Surbiton, Surrey KT6', country='UK')
        self.pq_uk_with_country_GB = PlaceQuery('32 Bond Road, Ste A, Surbiton, Surrey KT6', country='GB')

        # candidates
        self.good = Candidate(match_addr='123 Any St', locator='address', score=85.3)
        self.better = Candidate(match_addr='123 Any St', locator='parcel', score=92)
        self.best = Candidate(match_addr='123 Any St', locator='rooftop', score=100)
        self.wolf_good = Candidate(match_addr='1200 Callowhill St', locator='address', score=76)
        self.wolf_better = Candidate(match_addr='1200 Callowhill St', locator='parcel', score=90)
        self.wolf_best = Candidate(match_addr='1200 Callowhill St', locator='rooftop',
                                   score=99.9, x=-75.158, y=39.959)
        self.wolf_340 = Candidate(match_addr='340 N 12th St', locator='rooftop',
                                  score=99.5, x=-75.158, y=39.959)  # same coords
        self.inky = Candidate(match_addr='324 N Broad St', locator='rooftop',
                              score=99.9, x=-75.163, y=39.959)  # same y
        self.capt_thomas = Candidate(match_addr='843 Callowhill St', locator='rooftop',
                                     score=99.9, x=-75.163, y=39.959)  # same y
        self.reading_term = Candidate(match_addr='1200 Arch St', locator='rooftop',
                                      score=99.9, x=-75.163, y=39.953)  # same x
        self.with_address_types = Candidate(match_addr='123 Any St', locator='address',
                                            entity_types=['address', 'place'])
        self.with_nonsense_types = Candidate(match_addr='123 Any St', locator='address',
                                             entity_types=['house', 'building'])

        self.locators_worse_to_better = ['address', 'parcel', 'rooftop']

    def tearDown(self):
        pass

    def test_pro_country_CountryPreProcessor(self):
        """Test CountryPreProcessor"""
        acceptable_countries = ['US', 'UK']
        country_map = {'GB': 'UK'}  # 'from': 'to'
        place_in = self.pq_uk_with_country_GB
        place_out = CountryPreProcessor(acceptable_countries, country_map).process(place_in)
        country_exp = 'UK'
        self.assertEqual_(place_out.country, country_exp)

    def test_pro_country_RequireCountry(self):
        """Test RequireCountry preprocessor."""
        place_in = self.pq_us
        place_out = RequireCountry().process(place_in)
        place_exp = False
        self.assertEqual_(place_out, place_exp)

    def test_pro_CancelIfRegexInAttr(self):
        """Test CancelIfRegexInAttr preprocessor."""
        place_in = PlaceQuery('PO Box 123, Philadelphia, PA')
        place_out = CancelIfRegexInAttr(regex="po box", attrs=('query',)).process(place_in)
        place_exp = False
        self.assertEqual_(place_out, place_exp)

    def test_pro_CancelIfRegexInAttr_case_sensitive(self):
        """Test CancelIfRegexInAttr preprocessor using case-sensitive option."""
        place_in = PlaceQuery('PO Box 123, Philadelphia, PA')
        place_out = CancelIfRegexInAttr(regex="PO BOX", attrs=('query',),
                                        ignorecase=False).process(place_in)
        place_exp = place_in  # we should still have it because PO BOX does not match exactly
        self.assertEqual_(place_out, place_exp)

    def test_pro_CancelIfRegexInAttr_all_unicode(self):
        """Test CancelIfRegexInAttr preprocessor using unicode query, regex, and attrs."""
        place_in = PlaceQuery(u'PO Box 123, Lindström, MN')
        place_out = CancelIfRegexInAttr(regex=u"po box", attrs=(u'query',)).process(place_in)
        place_exp = False
        self.assertEqual_(place_out, place_exp)

    def test_pro_CancelIfRegexInAttr_unicode_query(self):
        """Test CancelIfRegexInAttr preprocessor using unicode query and default regex/attrs."""
        place_in = PlaceQuery(u'PO Box 123, Lindström, MN')
        place_out = CancelIfRegexInAttr(regex="po box", attrs=('query',)).process(place_in)
        place_exp = False
        self.assertEqual_(place_out, place_exp)

    def test_pro_CancelIfRegexInAttr_unicode_args(self):
        """Test CancelIfRegexInAttr preprocessor using bytestring query and unicode regex/attrs."""
        place_in = PlaceQuery(u'PO Box 123, Lindström, MN'.encode("utf-8"))
        # Test all 3 combinations for completeness, but only test the result once since if they
        # fail it will be by raising exceptions
        place_out = CancelIfRegexInAttr(regex=u"po box", attrs=('query',)).process(place_in)
        place_out = CancelIfRegexInAttr(regex="po box", attrs=(u'query',)).process(place_in)
        place_out = CancelIfRegexInAttr(regex=u"po box", attrs=('query',)).process(place_in)
        place_exp = False
        self.assertEqual_(place_out, place_exp)

    def test_pro_CancelIfPOBox(self):
        """Test CancelIfPOBox preprocessor."""
        place_in = PlaceQuery('PO Box 123, Philadelphia, PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='PO Box 123', city='Philadelphia', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='P.O Box 123', city='Philadelphia', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='P  O  box 123', city='Philadelphia', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='P.O. Box 123', city='Philadelphia', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='P.O. Box K', city='New Stanton', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='PO. Box K', city='New Stanton', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='P.O.B. 123', city='Philadelphia', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='P.O. BX123', city='Philadelphia', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery(address='POB 123', city='Philadelphia', state='PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery('POBOX 123, Philadelphia, PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, False)

        place_in = PlaceQuery('1200 Callowhill St, PO Box 466, Philadelphia, PA')
        place_out = CancelIfPOBox().process(place_in)
        self.assertEqual_(place_out, place_in)  # should still geocode because we a physical address

    def test_pro_filter_AttrFilter_exact(self):
        """Test AttrFilter postprocessor."""
        good_values = ['roof', 'parcel']
        candidates_in = [self.best, self.good, self.better]
        candidates_exp = [self.better]  # just the one with the parcel locator
        candidates_out = AttrFilter(good_values, 'locator', exact_match=True).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_filter_AttrFilter_inexact(self):
        """Test AttrFilter postprocessor with ``exact_match=False``."""
        good_values = ['roof', 'parcel']
        candidates_in = [self.best, self.good, self.better]
        candidates_exp = [self.best, self.better]  # roof is a substr of rooftop
        candidates_out = AttrFilter(good_values, 'locator', exact_match=False).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_filter_AttrExclude_exact(self):
        """Test AttrExclude with ``exact_match=True``."""
        bad_values = ['address', 'parc']
        candidates_in = [self.best, self.good, self.better]
        candidates_exp = [self.best, self.better]
        # The candidate with the 'parcel' locator stays because 'parcel' is not in bad values
        # and the processor by default only looks for exact matches against bad_values.
        candidates_out = AttrExclude(bad_values, 'locator', exact_match=True).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_filter_AttrExclude_inexact(self):
        """Test AttrExclude with ``exact_match=False``."""
        bad_values = ['address', 'parc']
        candidates_in = [self.best, self.good, self.better]
        candidates_exp = [self.best]
        # This can be confusing. There is only one because the match does NOT have to be exact, but
        # we are using this processor to EXCLUDE values, so an inexact match will result in fewer candidates
        candidates_out = AttrExclude(bad_values, 'locator', exact_match=False).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_postpro_GroupBy(self):
        """Test GroupBy postprocessor."""
        candidates_in = [self.best, self.good, self.better, self.wolf_best, self.wolf_good]
        candidates_exp = [self.best, self.wolf_best]
        candidates_out = GroupBy('match_addr').process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_postpro_GroupByMultiple(self):
        candidates_in = [self.wolf_best, self.wolf_340]
        candidates_exp = [self.wolf_best]
        candidates_out = GroupBy(('x', 'y')).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_parsing_ParseSingleLine(self):
        """Test ParseSingleLine preprocessor using single-line UK address."""
        place_in = PlaceQuery('32 Bond Road, Surbiton, Surrey KT6 7SH')
        place_out = ParseSingleLine().process(place_in)
        self.assertEqual_(place_out.address, '32 Bond Road')
        self.assertEqual_(place_out.city, 'Surbiton, Surrey')
        self.assertEqual_(place_out.postal, 'KT6 7SH')

    def test_pro_rename_AttrRename_inexact(self):
        """Test AttrRename postprocessor using partial search string."""
        candidates_in = [self.best]
        locator_exp = 'el_techo'
        candidates_out = AttrRename('locator', {'oofto': 'el_techo'}).process(candidates_in)
        self.assertEqual_(candidates_out[0].locator, locator_exp)

    def test_pro_rename_AttrRename_exact(self):
        """Test AttrRename postprocessor using exact search string."""
        candidates_in = [self.best]
        locator_exp = 'el_techo'
        candidates_out = AttrRename('locator', {'rooftop': 'el_techo'}).process(candidates_in)
        self.assertEqual_(candidates_out[0].locator, locator_exp)

    def test_pro_scoring_UseHighScoreIfAtLeast(self):
        """Test UseHighScoreIfAtLeast postprocessor."""
        candidates_in = [self.best, self.good, self.better]
        candidates_exp = [self.best, self.better]
        candidates_out = UseHighScoreIfAtLeast(90).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_scoring_ScoreSorter(self):
        """Test ScoreSorter postprocessor."""
        candidates_in = [self.best, self.good, self.better]
        candidates_exp = [self.best, self.better, self.good]
        candidates_out = ScoreSorter().process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_scoring_ScoreSorter_asc(self):
        """Test ScoreSorter postprocessor with ``reverse=False``."""
        candidates_in = [self.best, self.good, self.better]
        candidates_exp = [self.good, self.better, self.best]
        candidates_out = ScoreSorter(reverse=False).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_sort_AttrSorter(self):
        """Test AttrSorter postprocessor."""
        candidates_in = [self.better, self.best, self.good]
        candidates_exp = [self.good, self.better, self.best]
        candidates_out = AttrSorter(self.locators_worse_to_better).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_sort_AttrReverseSorter(self):
        """Test AttrReverseSorter postprocessor."""
        candidates_in = [self.better, self.best, self.good]
        candidates_exp = [self.best, self.better, self.good]  # reverse order of self.locators_worse_to_better
        candidates_out = AttrReverseSorter(self.locators_worse_to_better).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_streetnumber_ReplaceRangeWithNumber(self):
        """Test ReplaceRangeWithNumber preprocessor."""
        place_in = PlaceQuery('4452-54 Main Street, Philadelphia')  # Mom's Pizza in Manayunk
        place_out = ReplaceRangeWithNumber().process(place_in)
        query_exp = '4452 Main Street, Philadelphia'
        self.assertEqual_(place_out.query, query_exp)

        zip_plus_4 = '19127-1112'
        place_in = PlaceQuery(zip_plus_4)  # sets PlaceQuery.query to zip_plus_4 on init
        place_out = ReplaceRangeWithNumber().process(place_in)
        self.assertEqual_(place_out.query, zip_plus_4)

    def test_pro_SnapPoints(self):
        """Take two candidates within 50 metres and eliminate one."""
        candidates_in = [Candidate(match_addr='340 N 12th St, Philadelphia, PA, 19107',
                                   x=-75.158433167, y=39.958727992),
                         Candidate(match_addr='1200 Callowhill St, Philadelphia, PA, 19123',
                                   x=-75.158303781, y=39.959040684)]  # about 40m away
        candidates_exp = [candidates_in[0]]  # should just keep the first one.
        candidates_out = SnapPoints(distance=50).process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_filter_AttrListIncludes(self):
        """Test AttrListIncludes postprocessor."""
        good_values = ['address']
        candidates_in = [self.with_address_types, self.with_nonsense_types]
        candidates_exp = [self.with_address_types]
        candidates_out = AttrListIncludes(good_values, 'entity_types').process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)

    def test_pro_filter_AttrListExcludes(self):
        """Test AttrListExcludes postprocessor."""
        bad_values = ['house']
        candidates_in = [self.with_address_types, self.with_nonsense_types]
        candidates_exp = [self.with_address_types]
        candidates_out = AttrListExcludes(bad_values, 'entity_types').process(candidates_in)
        self.assertEqual_(candidates_out, candidates_exp)


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout)
    logging.getLogger(__name__).setLevel(logging.DEBUG)
    unittest.main()



class _Processor():
    def _init_helper(self, vars_):
        """Overwrite defaults (if they exist) with arguments passed to constructor"""
        for k in vars_:
            if k == 'kwargs':
                for kwarg in vars_[k]:
                    setattr(self, kwarg, vars_[k][kwarg])
            elif k != 'self':
                setattr(self, k, vars_[k])

    def __init__(self, **kwargs):
        """
        In a subclass, arguments may be formally defined to avoid the use of keywords
        (and to throw errors when bogus keyword arguments are passed)::

            def __init__(self, arg1='foo', arg2='bar')

        """
        self._init_helper(vars())



#!/usr/bin/env python

import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="python-omgeo",
    version="6.1.0",
    description="Geocoding Library using ESRI, Google, Bing Maps, US Census, OpenStreetMap, Pelias, and MapQuest geocoders",
    author="Azavea, Inc.",
    author_email="info@azavea.com",
    url="http://github.com/azavea/python-omgeo",
    license="MIT",
    long_description=read('README.rst'),
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: GIS',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 3'
    ],
    dependency_links=[],
    install_requires=[
        'requests >= 2.18',
    ],
    test_suite='omgeo.tests.tests',
)



from .base import GeocodeService
import logging
from omgeo.places import Candidate
from omgeo.preprocessors import ReplaceRangeWithNumber
from omgeo.postprocessors import AttrFilter, AttrExclude

logger = logging.getLogger(__name__)


class Nominatim(GeocodeService):
    """
    Class to geocode using `Nominatim services hosted
    by MapQuest <http://open.mapquestapi.com/nominatim/>`_.
    """
    _wkid = 4326
    _endpoint = 'http://open.mapquestapi.com/nominatim/v1/search'

    DEFAULT_ACCEPTED_ENTITIES = ['building.', 'historic.castle', 'leisure.ice_rink',
                                 'leisure.miniature_golf',
                                 'leisure.sports_centre', 'lesiure.stadium', 'leisure.track',
                                 'lesiure.water_park', 'man_made.lighthouse', 'man_made.works',
                                 'military.barracks', 'military.bunker', 'office.', 'place.house',
                                 'amenity.', 'power.generator', 'railway.station',
                                 'shop.', 'tourism.']

    DEFAULT_REJECTED_ENTITIES = ['amenity.drinking_water',
                                 'amentity.bicycle_parking', 'amentity.ev_charging',
                                 'amentity.grit_bin', 'amentity.atm',
                                 'amentity.hunting_stand', 'amentity.post_box']

    DEFAULT_PREPROCESSORS = [ReplaceRangeWithNumber()]  # 766-68 Any St. -> 766 Any St.
    """Preprocessors to use with this geocoder service, in order of desired execution."""

    DEFAULT_POSTPROCESSORS = [
        AttrFilter(DEFAULT_ACCEPTED_ENTITIES, 'entity', exact_match=False),
        AttrExclude(DEFAULT_REJECTED_ENTITIES, 'entity')
    ]
    """Postprocessors to use with this geocoder service, in order of desired execution."""

    def __init__(self, preprocessors=None, postprocessors=None, settings=None):
        preprocessors = Nominatim.DEFAULT_PREPROCESSORS if preprocessors is None else preprocessors
        postprocessors = Nominatim.DEFAULT_POSTPROCESSORS if postprocessors is None else postprocessors
        GeocodeService.__init__(self, preprocessors, postprocessors, settings)

    def _geocode(self, pq):
        query = {'q': pq.query,
                 'countrycodes': pq.country,  # only takes ISO-2
                 'format': 'json'}

        if pq.viewbox is not None:
            query = dict(query, **{'viewbox': pq.viewbox.to_mapquest_str(), 'bounded': pq.bounded})

        response_obj = self._get_json_obj(self._endpoint, query)

        returned_candidates = []  # this will be the list returned
        for r in response_obj:
            c = Candidate()
            c.locator = 'parcel'  # we don't have one but this is the closest match
            c.entity = '%s.%s' % (r['class'], r['type'])  # ex.: "place.house"
            c.match_addr = r['display_name']  # ex. "Wolf Building, 340, N 12th St, Philadelphia, Philadelphia County, Pennsylvania, 19107, United States of America" #TODO: shorten w/ pieces
            c.x = float(r['lon'])  # long, ex. -122.13 # cast to float in 1.3.4
            c.y = float(r['lat'])  # lat, ex. 47.64 # cast to float in 1.3.4
            c.wkid = self._wkid
            c.geoservice = self.__class__.__name__
            returned_candidates.append(c)
        return returned_candidates



import copy
from datetime import datetime
from json import loads
import logging
from traceback import format_exc
from xml.dom import minidom

import requests


logger = logging.getLogger(__name__)


class UpstreamResponseInfo():
    """
    Description of API call result from an upstream provider.
    For cleaning and consistency, set attributes using the given methods
    (the constructor will automatically use these methods, as well).
    """

    def set_response_code(self, response_code):
        if response_code is not None and type(response_code) is not int:
            raise Exception('response_code must be an integer.')
        else:
            self.response_code = response_code

    def set_response_time(self, response_time):
        if response_time is not None:
            if type(response_time) not in (int, float):
                raise Exception('If response_time is provided,'
                                'it must be an integer or float.')
            elif response_time < 0:
                raise Exception('response_time cannot be negative.')
            self.response_time = int(round(response_time))
        else:
            self.response_time = None

    def set_success(self, success):
        if type(success) is not bool:
            raise ('success must be a boolean value.')
        else:
            self.success = success

    def __init__(self, geoservice, processed_pq, response_code=None, response_time=None,
                 success=True, errors=None):
        """
        :arg str geoservice: name of the upstream provider used (required)
        :arg PlaceQuery processed_pq: Processed PlaceQuery object (required)
        :arg response_code: HTTP response code (default None)
        :arg response_time: time in milliseconds that it takes to get a
                            response (default None)
        :arg bool success: indicates if the API call was successful. A 200 response
                           with no candidates is still considered a success.
                           (default True)
        :arg list errors: human-readable error descriptions
        """

        if errors is None:
            errors = []
        self.geoservice = geoservice
        self.processed_pq = processed_pq
        self.set_response_code(response_code)
        self.set_response_time(response_time)
        self.set_success(success)
        self.errors = errors

    def __repr__(self):
        if self.response_code is None:
            repr_ = '%s %sms' % (self.geoservice, self.response_time)
        else:
            repr_ = '%s %s %sms' % (self.geoservice, self.response_code, self.response_time)
        return '<%s>' % repr_


class GeocodeService():
    """
    Base class for geocoding API wrappers
    """

    #: API base endpoint URL to use
    _endpoint = ''

    def __init__(self, preprocessors=None, postprocessors=None,
                 settings=None):
        """
        Overwrite _preprocessors, _postprocessors, and _settings
        if they are set.
        """

        #: Preprocessor classes to apply to the given PlaceQuery, usually
        #: overwritten in subclass.
        self._preprocessors = []

        #: Postprocessor classes to apply to the list of Candidates obtained,
        #: usually overwritten in subclass.
        self._postprocessors = []

        #: Settings for this geocoder, usually overwritten by subclass
        self._settings = {}

        if preprocessors is not None:
            self._preprocessors = preprocessors
        if postprocessors is not None:
            self._postprocessors = postprocessors
        if settings is not None:
            for key in settings:
                self._settings[key] = settings[key]

    def _settings_checker(self, required_settings=None, accept_none=True):
        """
        Take a list of required _settings dictionary keys
        and make sure they are set. This can be added to a custom
        constructor in a subclass and tested to see if it returns ``True``.

        :arg list required_settings: A list of required keys to look for.
        :arg bool accept_none: Boolean set to True if None is an acceptable
                               setting. Set to False if None is not an
                               acceptable setting.

        :returns: * bool ``True`` if all required settings exist, OR
                  * str <key name> for the first key missing from _settings.
        """
        if required_settings is not None:
            for keyname in required_settings:
                if keyname not in self._settings:
                    return keyname
                if accept_none is False and self._settings[keyname] is None:
                    return keyname
        return True

    def _get_response(self, endpoint, query, is_post=False):
        """Returns response or False in event of failure"""
        timeout_secs = self._settings.get('timeout', 10)
        headers = self._settings.get('request_headers', {})
        try:
            if is_post:
                response = requests.post(
                    endpoint, data=query, headers=headers, timeout=timeout_secs)
            else:
                response = requests.get(
                    endpoint, params=query, headers=headers, timeout=timeout_secs)
        except requests.exceptions.Timeout:
            raise Exception('API request timed out after %s seconds.' % timeout_secs)
        except Exception as e:
            raise e

        if response.status_code != 200:
            raise Exception('Received status code %s from %s. Content is:\n%s'
                            % (response.status_code,
                               self.get_service_name(),
                               response.text))
        return response

    def _get_json_obj(self, endpoint, query, is_post=False):
        """
        Return False if connection could not be made.
        Otherwise, return a response object from JSON.
        """
        response = self._get_response(endpoint, query, is_post=is_post)
        content = response.text
        try:
            return loads(content)
        except ValueError:
            raise Exception('Could not decode content to JSON:\n%s'
                            % self.__class__.__name__, content)

    def _get_xml_doc(self, endpoint, query, is_post=False):
        """
        Return False if connection could not be made.
        Otherwise, return a minidom Document.
        """
        response = self._get_response(endpoint, query, is_post=is_post)
        return minidom.parse(response.text)

    def _geocode(self, place_query):
        """
        Given a (preprocessed) PlaceQuery object,
        return a list of of Candidate objects.
        """
        raise NotImplementedError(
            'GeocodeService subclasses must implement _geocode().')

    def geocode(self, pq):
        """
        :arg PlaceQuery pq: PlaceQuery instance
        :rtype: tuple
        :returns: post-processed list of Candidate objects and
                  and UpstreamResponseInfo object if an API call was made.

                  Examples:

                  Preprocessor throws out request::

                      ([], None)

                  Postprocessor throws out some candidates::

                      ([<Candidate obj>, <Candidate obj>, ...], <UpstreamResponseInfo obj>)

                  Postprocessor throws out all candidates::

                      ([], <UpstreamResponseInfo obj>)

                  An exception occurs while making the API call::

                      ([], <UpstreamResponseInfo obj>)

        """
        processed_pq = copy.copy(pq)
        for p in self._preprocessors:
            processed_pq = p.process(processed_pq)
            if not processed_pq:
                return [], None
        upstream_response_info = UpstreamResponseInfo(self.get_service_name(),
                                                      processed_pq)
        try:
            start = datetime.now()
            candidates = self._geocode(processed_pq)
            end = datetime.now()
            response_time_sec = (end - start).total_seconds()
            upstream_response_info.set_response_time(1000 * response_time_sec)
        except Exception:
            upstream_response_info.set_success(False)
            upstream_response_info.errors.append(format_exc())
            return [], upstream_response_info
        if len(candidates) > 0:
            for p in self._postprocessors:  # apply universal candidate postprocessing
                candidates = p.process(candidates)  # merge lists
        return candidates, upstream_response_info

    def get_service_name(self):
        return self.__class__.__name__



# flake8: noqa: F401
from .bing import Bing
from .esri import EsriWGS
from .us_census import USCensus
from .mapquest import MapQuest, MapQuestSSL
from .nominatim import Nominatim
from .pelias import Pelias
from .google import Google



from omgeo.processor import _Processor
import re

# Python 2/3 compatibility: set up 'unicode' for use in string type checking
try:
    unicode
except NameError:
    unicode = str


class _PreProcessor(_Processor):
    """Takes, processes, and returns a geocoding.places.PlaceQuery object."""
    def process(self, pq):
        raise NotImplementedError(
            'PreProcessor subclasses must implement process().')


class ReplaceRangeWithNumber(_PreProcessor):
    """
    Class to take only the first part of an address range
    or hyphenated house number to use for geocoding.

    This affects the query and address PlaceQuery attributes.

    =============================== ========================================
    Input                           Output
    =============================== ========================================
    ``4109-4113 Main St``           ``4109 Main St``
    ``4109-13 Main St``             ``4109 Main St``
    ``322-1/2 Water Street``        ``322 Water Street``
    ``123-2 Maple Lane``            ``123 Maple Lane``
    ``272-B Greenough St, 19127``   ``272 Greenough St, 19127``
    ``272 Greenough St 19127-1112`` ``272 Greenough St 19127-1112``
    ``19127-1112``                  ``19127-1112`` (not affected)
    ``76-20 34th Ave, Queens NY``   ``76 34th Ave, Queens NY`` (see warning)
    =============================== ========================================

    .. warning::

       This may cause problems with addresses presented in the
       hyphenated Queens-style format, where the part before the
       hyphen indicates the cross street, and the part after
       indicates the house number.
    """

    #: Regular expression to represent ranges like:
    #:  * 789-791
    #:  * 789-91
    #:  * 201A-201B
    #:  * 201A-B
    RE_STREET_NUMBER = re.compile(r'(^\d+\w*-\d*\w*)\s', re.IGNORECASE)

    def replace_range(self, addr_str):
        match = self.RE_STREET_NUMBER.match(addr_str)
        if match is not None:
            old = match.group(1)
            new = old.split('-', 1)[0]
            addr_str = addr_str.replace(old, new, 1)
        return addr_str

    def process(self, pq):
        """
        :arg PlaceQuery pq: PlaceQuery instance
        :returns: PlaceQuery instance with truncated address range / number
        """
        pq.query = self.replace_range(pq.query)
        pq.address = self.replace_range(pq.address)
        return pq


class ParseSingleLine(_PreProcessor):
    """
    Adapted from `Cicero Live <http://azavea.com/packages/azavea_cicero/blocks/cicero_live/view.js>`_
    """
    # Some Regexes:
    re_unit_numbered = re.compile(r'(su?i?te|p\W*[om]\W*b(?:ox)?|(?:ap|dep)(?:ar)?t(?:me?nt)?|ro*m|flo*r?|uni?t|bu?i?ldi?n?g|ha?nga?r|lo?t|pier|slip|spa?ce?|stop|tra?i?le?r|bo?x|no\.?)\s+|#', re.IGNORECASE)
    re_unit_not_numbered = re.compile(r'ba?se?me?n?t|fro?nt|lo?bby|lowe?r|off?i?ce?|pe?n?t?ho?u?s?e?|rear|side|uppe?r', re.IGNORECASE)
    re_UK_postcode = re.compile(r'[A-Z]{1,2}[0-9R][0-9A-Z]? *[0-9][A-Z]{0,2}', re.IGNORECASE)
    re_blank = re.compile(r'\s')

    def _comma_join(self, left, right):
        if left == '':
            return right
        else:
            return '%s, %s' % (left, right)

    def process(self, pq):
        """
        :arg PlaceQuery pq: PlaceQuery instance
        :returns: PlaceQuery instance with :py:attr:`query`
                  converted to individual elements
        """
        if pq.query != '':
            postcode = address = city = ''  # define the vars we'll use

            # global regex postcode search, pop off last result
            postcode_matches = self.re_UK_postcode.findall(pq.query)
            if len(postcode_matches) > 0:
                postcode = postcode_matches[-1]

            query_parts = [part.strip() for part in pq.query.split(',')]

            if postcode != '' and re.search(postcode, query_parts[0]):
                # if postcode is in the first part of query_parts, there are probably no commas
                # get just the part before the postcode
                part_before_postcode = query_parts[0].split(postcode)[0].strip()
                if self.re_blank.search(part_before_postcode) is None:
                    address = part_before_postcode
                else:
                    address = query_parts[0]  # perhaps it isn't really a postcode (apt num, etc)
            else:
                address = query_parts[0]  # no postcode to worry about

            for part in query_parts[1:]:
                part = part.strip()
                if postcode != '' and re.search(postcode, part) is not None:
                    part = part.replace(postcode, '').strip()  # if postcode is in part, remove it

                if self.re_unit_numbered.search(part) is not None:
                    # test to see if part is secondary address, like "Ste 402"
                    address = self._comma_join(address, part)
                elif self.re_unit_not_numbered.search(part) is not None:
                    # ! might cause problems if 'Lower' or 'Upper' is in the city name
                    # test to see if part is secondary address, like "Basement"
                    address = self._comma_join(address, part)
                else:
                    city = self._comma_join(city, part)  # it's probably a city (or "City, County")
            # set pq parts if they aren't already set (we don't want to overwrite explicit params)
            pq.postal = pq.postal or postcode
            pq.address = pq.address or address
            pq.city = pq.city or city

        return pq


class ComposeSingleLine(_PreProcessor):
    """ Compose address components into a single-line query if no query is already defined. """
    def process(self, pq):
        if pq.query == '':
            parts = [pq.address, pq.city, pq.subregion]
            parts.append(' '.join([p for p in (pq.state, pq.postal) if p != '']))
            if pq.country != '':
                parts.append(pq.country)
            pq.query = ', '.join([part for part in parts if part != ''])

        return pq


class CountryPreProcessor(_PreProcessor):
    """
    Used to filter acceptable countries
    and standardize country names or codes.
    """

    def __init__(self, acceptable_countries=None, country_map=None):
        """
        :arg list acceptable_countries: A list of acceptable countries.
                                        None is used to indicate that all countries are acceptable.
                                        (default ``[]``)

                                        An empty string is also an acceptable country. To require
                                        a country, use the `RequireCountry` preprocessor.

        :arg dict country_map: A map of the input PlaceQuery.country property
                               to the country value accepted by the geocoding service.

                               For example, suppose that the geocoding service recognizes
                               'GB', but not 'UK' -- and 'US', but not 'USA'::

                                    country_map = {'UK':'GB', 'USA':'US'}

        """
        self.acceptable_countries = acceptable_countries if acceptable_countries is not None else []
        self.country_map = country_map if country_map is not None else {}

    def process(self, pq):
        """
        :arg PlaceQuery pq: PlaceQuery instance
        :returns: modified PlaceQuery, or ``False`` if country is not acceptable.
        """
        # Map country, but don't let map overwrite
        if pq.country not in self.acceptable_countries and pq.country in self.country_map:
            pq.country = self.country_map[pq.country]
        if pq.country != '' and \
           self.acceptable_countries != [] and \
           pq.country not in self.acceptable_countries:
            return False
        return pq

    def __repr__(self):
        return '<%s: Accept %s mapped as %s>' % (self.__class__.__name__,
                                                 self.acceptable_countries, self.country_map)


class CancelIfRegexInAttr(_PreProcessor):
    """
    Return False if given regex is found in ANY of the given
    PlaceQuery attributes, otherwise return original PlaceQuery instance.
    In the event that a given attribute does not exist in the given
    PlaceQuery, no exception will be raised.
    """
    def __init__(self, regex, attrs, ignorecase=True):
        """
        :arg str regex: a regex string to match (represents what you do *not* want)
        :arg attrs: a list or tuple of strings of attribute names to look through
        :arg bool ignorecase: set to ``False`` for a case-sensitive match (default ``True``)
        """
        regex_type = type(regex)
        if type(regex) not in (str, unicode):
            raise Exception('First param "regex" must be a regex of type'
                            ' str or unicode, not %s.' % regex_type)
        attrs_type = type(attrs)
        if attrs_type not in (list, tuple):
            raise Exception('Second param "attrs" must be a list or tuple'
                            ' of PlaceQuery attributes, not %s.' % attrs_type)
        if any(type(attr) not in (str, unicode) for attr in attrs):
            raise Exception('All given PlaceQuery attributes must be strings.')
        self.attrs = attrs
        if ignorecase:
            self.regex = re.compile(regex, re.IGNORECASE)
        else:
            self.regex = re.compile(regex)

    def process(self, pq):
        attrs = [getattr(pq, attr) for attr in self.attrs if hasattr(pq, attr)]
        for attr in attrs:
            # Handle string type mismatch by decoding if the first attempt raises TypeError
            try:
                match = self.regex.match(attr)
            except TypeError:
                match = self.regex.match(attr.decode("utf-8"))
            if match is not None:
                return False  # don't return the placequery if any match is found
        return pq

    def __repr__(self):
        case_sensitive = 'insensitive' if self.ignorecase else 'sensitive'
        return '<%s: Break if %s in %s (case %s)>' % (self.__class__.__name__,
                                                      self.regex, self.attrs, case_sensitive)


class CancelIfPOBox(_PreProcessor):
    def process(self, pq):
        """
        :arg PlaceQuery pq: PlaceQuery instance
        :returns: ``False`` if the address is starts with any variation of "PO Box".
                  Otherwise, return original :py:class:`PlaceQuery`.
        """
        regex = r'^\s*P\.?\s*O\.?\s*B\.?O?X?[\s\d]'
        return CancelIfRegexInAttr(regex, ('address', 'query')).process(pq)


class RequireCountry(_PreProcessor):
    """
    Return False if no default country is set in first parameter.
    Otherwise, return the default country if country is empty.
    """
    def __init__(self, default_country=''):
        """
        :arg str default_country: default country to use if there is
                                  no country set in the PlaceQuery instance sent to this processor.
                                  If this argument is not set or empty and PlaceQuery instance does
                                  not have a country (pq.country == ''), the processor will return
                                  False and the PlaceQuery will be rejected during geocoding.
                                  (default ``''``)

        """
        self.default_country = default_country

    def process(self, pq):
        """
        :arg PlaceQuery pq: PlaceQuery instance
        :returns: One of the three following values:
                   * unmodified PlaceQuery instance if pq.country is not empty
                   * PlaceQuery instance with pq.country changed to default country.
                   * ``False`` if pq.country is empty and self.default_country == ''.
        """
        if pq.country.strip() == '':
            if self.default_country == '':
                return False
            else:
                pq.country = self.default_country
        return pq
